@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2015 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2015 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -29,13 +29,15 @@ int	alerts_enabled(struct window *, int);
 void	alerts_callback(int, short, void *);
 void	alerts_reset(struct window *);
 
+void	alerts_run_hook(struct session *, struct winlink *, int);
+int	alerts_check_all(struct session *, struct winlink *);
 int	alerts_check_bell(struct session *, struct winlink *);
 int	alerts_check_activity(struct session *, struct winlink *);
 int	alerts_check_silence(struct session *, struct winlink *);
 void	alerts_ring_bell(struct session *);
 
 void
-alerts_timer(unused int fd, unused short events, void *arg)
+alerts_timer(__unused int fd, __unused short events, void *arg)
 {
 	struct window	*w = arg;
 
@@ -45,7 +47,7 @@ alerts_timer(unused int fd, unused short events, void *arg)
 }
 
 void
-alerts_callback(unused int fd, unused short events, unused void *arg)
+alerts_callback(__unused int fd, __unused short events, __unused void *arg)
 {
 	struct window	*w;
 	struct session	*s;
@@ -59,11 +61,7 @@ alerts_callback(unused int fd, unused short events, unused void *arg)
 					continue;
 				flags = w->flags;
 
-				alerts  = alerts_check_bell(s, wl);
-				alerts |= alerts_check_activity(s, wl);
-				alerts |= alerts_check_silence(s, wl);
-				if (alerts != 0)
-					server_status_session(s);
+				alerts = alerts_check_all(s, wl);
 
 				log_debug("%s:%d @%u alerts check, alerts %#x, "
 				    "flags %#x", s->name, wl->idx, w->id,
@@ -74,25 +72,58 @@ alerts_callback(unused int fd, unused short events, unused void *arg)
 	alerts_fired = 0;
 }
 
+void
+alerts_run_hook(struct session *s, struct winlink *wl, int flags)
+{
+	struct cmd_find_state	 fs;
+
+	if (cmd_find_from_winlink(&fs, s, wl) != 0)
+		return;
+
+	if (flags & WINDOW_BELL)
+		hooks_run(s->hooks, NULL, &fs, "alert-bell");
+	if (flags & WINDOW_SILENCE)
+		hooks_run(s->hooks, NULL, &fs, "alert-silence");
+	if (flags & WINDOW_ACTIVITY)
+		hooks_run(s->hooks, NULL, &fs, "alert-activity");
+}
+
+int
+alerts_check_all(struct session *s, struct winlink *wl)
+{
+	int	alerts;
+
+	alerts  = alerts_check_bell(s, wl);
+	alerts |= alerts_check_activity(s, wl);
+	alerts |= alerts_check_silence(s, wl);
+	if (alerts != 0) {
+		alerts_run_hook(s, wl, alerts);
+		server_status_session(s);
+	}
+
+	return (alerts);
+}
+
+void
+alerts_check_session(struct session *s)
+{
+	struct winlink	*wl;
+
+	RB_FOREACH(wl, winlinks, &s->windows)
+		alerts_check_all(s, wl);
+}
+
 int
 alerts_enabled(struct window *w, int flags)
 {
-	struct session	*s;
-
+	if (flags & WINDOW_BELL)
+		return (1);
 	if (flags & WINDOW_ACTIVITY) {
-		if (options_get_number(&w->options, "monitor-activity"))
+		if (options_get_number(w->options, "monitor-activity"))
 			return (1);
 	}
 	if (flags & WINDOW_SILENCE) {
-		if (options_get_number(&w->options, "monitor-silence") != 0)
-			return (1);
-	}
-	if (~flags & WINDOW_BELL)
-		return (0);
-	RB_FOREACH(s, sessions, &sessions) {
-		if (!session_has(s, w))
-			continue;
-		if (options_get_number(&s->options, "bell-action") != BELL_NONE)
+		if (options_get_number(w->options, "monitor-silence") != 0)
 			return (1);
 	}
 	return (0);
@@ -116,7 +147,7 @@ alerts_reset(struct window *w)
 	event_del(&w->alerts_timer);
 
 	timerclear(&tv);
-	tv.tv_sec = options_get_number(&w->options, "monitor-silence");
+	tv.tv_sec = options_get_number(w->options, "monitor-silence");
 
 	log_debug("@%u alerts timer reset %u", w->id, (u_int)tv.tv_sec);
 	if (tv.tv_sec != 0)
@@ -132,10 +163,10 @@ alerts_queue(struct window *w, int flags)
 	if (!event_initialized(&w->alerts_timer))
 		evtimer_set(&w->alerts_timer, alerts_timer, w);
 
-	if (w->flags & flags)
-		return;
-	w->flags |= flags;
-	log_debug("@%u alerts flags added %#x", w->id, flags);
+	if ((w->flags & flags) != flags) {
+		w->flags |= flags;
+		log_debug("@%u alerts flags added %#x", w->id, flags);
+	}
 
 	if (!alerts_fired && alerts_enabled(w, flags)) {
 		log_debug("alerts check queued (by @%u)", w->id);
@@ -151,20 +182,20 @@ alerts_check_bell(struct session *s, struct winlink *wl)
 	struct window	*w = wl->window;
 	int		 action, visual;
 
-	if (!(w->flags & WINDOW_BELL) || wl->flags & WINLINK_BELL)
+	if (!(w->flags & WINDOW_BELL))
 		return (0);
-	if (s->curw != wl || s->flags & SESSION_UNATTACHED)
+	if (s->curw != wl) {
 		wl->flags |= WINLINK_BELL;
-	if (s->flags & SESSION_UNATTACHED)
-		return (0);
+		w->flags &= ~WINDOW_BELL;
+	}
 	if (s->curw->window == w)
 		w->flags &= ~WINDOW_BELL;
 
-	action = options_get_number(&s->options, "bell-action");
+	action = options_get_number(s->options, "bell-action");
 	if (action == BELL_NONE)
 		return (0);
 
-	visual = options_get_number(&s->options, "visual-bell");
+	visual = options_get_number(s->options, "visual-bell");
 	TAILQ_FOREACH(c, &clients, entry) {
 		if (c->session != s || c->flags & CLIENT_CONTROL)
 			continue;
@@ -198,17 +229,17 @@ alerts_check_activity(struct session *s, struct winlink *wl)
 
 	if (!(w->flags & WINDOW_ACTIVITY) || wl->flags & WINLINK_ACTIVITY)
 		return (0);
-	if (s->curw == wl && !(s->flags & SESSION_UNATTACHED))
+	if (s->curw == wl)
 		return (0);
 
-	if (!options_get_number(&w->options, "monitor-activity"))
+	if (!options_get_number(w->options, "monitor-activity"))
 		return (0);
 
-	if (options_get_number(&s->options, "bell-on-alert"))
+	if (options_get_number(s->options, "bell-on-alert"))
 		alerts_ring_bell(s);
 	wl->flags |= WINLINK_ACTIVITY;
 
-	if (options_get_number(&s->options, "visual-activity")) {
+	if (options_get_number(s->options, "visual-activity")) {
 		TAILQ_FOREACH(c, &clients, entry) {
 			if (c->session != s)
 				continue;
@@ -230,17 +261,17 @@ alerts_check_silence(struct session *s, struct winlink *wl)
 
 	if (!(w->flags & WINDOW_SILENCE) || wl->flags & WINLINK_SILENCE)
 		return (0);
-	if (s->curw == wl && !(s->flags & SESSION_UNATTACHED))
+	if (s->curw == wl)
 		return (0);
 
-	if (options_get_number(&w->options, "monitor-silence") == 0)
+	if (options_get_number(w->options, "monitor-silence") == 0)
 		return (0);
 
-	if (options_get_number(&s->options, "bell-on-alert"))
+	if (options_get_number(s->options, "bell-on-alert"))
 		alerts_ring_bell(s);
 	wl->flags |= WINLINK_SILENCE;
 
-	if (options_get_number(&s->options, "visual-silence")) {
+	if (options_get_number(s->options, "visual-silence")) {
 		TAILQ_FOREACH(c, &clients, entry) {
 			if (c->session != s)
 				continue;
