@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -29,14 +29,19 @@
 
 enum cmd_retval	 cmd_bind_key_exec(struct cmd *, struct cmd_q *);
 
-enum cmd_retval	 cmd_bind_key_mode_table(struct cmd *, struct cmd_q *, int);
+enum cmd_retval	 cmd_bind_key_mode_table(struct cmd *, struct cmd_q *,
+		     key_code);
 
 const struct cmd_entry cmd_bind_key_entry = {
-	"bind-key", "bind",
-	"cnrt:", 1, -1,
-	"[-cnr] [-t mode-table] key command [arguments]",
-	0,
-	cmd_bind_key_exec
+	.name = "bind-key",
+	.alias = "bind",
+
+	.args = { "cnrR:t:T:", 1, -1 },
+	.usage = "[-cnr] [-t mode-table] [-R repeat-count] [-T key-table] key "
+	         "command [arguments]",
+
+	.flags = 0,
+	.exec = cmd_bind_key_exec
 };
 
 enum cmd_retval
@@ -45,7 +50,8 @@ cmd_bind_key_exec(struct cmd *self, struct cmd_q *cmdq)
 	struct args	*args = self->args;
 	char		*cause;
 	struct cmd_list	*cmdlist;
-	int		 key;
+	key_code	 key;
+	const char	*tablename;
 
 	if (args_has(args, 't')) {
 		if (args->argc != 2 && args->argc != 3) {
@@ -60,7 +66,7 @@ cmd_bind_key_exec(struct cmd *self, struct cmd_q *cmdq)
 	}
 
 	key = key_string_lookup_string(args->argv[0]);
-	if (key == KEYC_NONE) {
+	if (key == KEYC_NONE || key == KEYC_UNKNOWN) {
 		cmdq_error(cmdq, "unknown key: %s", args->argv[0]);
 		return (CMD_RETURN_ERROR);
 	}
@@ -68,6 +74,13 @@ cmd_bind_key_exec(struct cmd *self, struct cmd_q *cmdq)
 	if (args_has(args, 't'))
 		return (cmd_bind_key_mode_table(self, cmdq, key));
 
+	if (args_has(args, 'T'))
+		tablename = args_get(args, 'T');
+	else if (args_has(args, 'n'))
+		tablename = "root";
+	else
+		tablename = "prefix";
+
 	cmdlist = cmd_list_parse(args->argc - 1, args->argv + 1, NULL, 0,
 	    &cause);
 	if (cmdlist == NULL) {
@@ -76,21 +89,20 @@ cmd_bind_key_exec(struct cmd *self, struct cmd_q *cmdq)
 		return (CMD_RETURN_ERROR);
 	}
 
-	if (!args_has(args, 'n'))
-	    key |= KEYC_PREFIX;
-	key_bindings_add(key, args_has(args, 'r'), cmdlist);
+	key_bindings_add(tablename, key, args_has(args, 'r'), cmdlist);
 	return (CMD_RETURN_NORMAL);
 }
 
 enum cmd_retval
-cmd_bind_key_mode_table(struct cmd *self, struct cmd_q *cmdq, int key)
+cmd_bind_key_mode_table(struct cmd *self, struct cmd_q *cmdq, key_code key)
 {
 	struct args			*args = self->args;
-	const char			*tablename;
+	const char			*tablename, *arg;
 	const struct mode_key_table	*mtab;
 	struct mode_key_binding		*mbind, mtmp;
 	enum mode_key_cmd		 cmd;
-	const char			*arg;
+	char				*cause;
+	u_int				 repeat;
 
 	tablename = args_get(args, 't');
 	if ((mtab = mode_key_findtable(tablename)) == NULL) {
@@ -104,18 +116,44 @@ cmd_bind_key_mode_table(struct cmd *self, struct cmd_q *cmdq, int key)
 		return (CMD_RETURN_ERROR);
 	}
 
-	if (cmd != MODEKEYCOPY_COPYPIPE) {
-		if (args->argc != 2) {
-			cmdq_error(cmdq, "no argument allowed");
-			return (CMD_RETURN_ERROR);
+	switch (cmd) {
+	case MODEKEYCOPY_APPENDSELECTION:
+	case MODEKEYCOPY_COPYSELECTION:
+	case MODEKEYCOPY_STARTNAMEDBUFFER:
+		if (args->argc == 2)
+			arg = NULL;
+		else {
+			arg = args->argv[2];
+			if (strcmp(arg, "-x") != 0) {
+				cmdq_error(cmdq, "unknown argument");
+				return (CMD_RETURN_ERROR);
+			}
 		}
-		arg = NULL;
-	} else {
+		break;
+	case MODEKEYCOPY_COPYPIPE:
 		if (args->argc != 3) {
 			cmdq_error(cmdq, "no argument given");
 			return (CMD_RETURN_ERROR);
 		}
 		arg = args->argv[2];
+		break;
+	default:
+		if (args->argc != 2) {
+			cmdq_error(cmdq, "no argument allowed");
+			return (CMD_RETURN_ERROR);
+		}
+		arg = NULL;
+		break;
+	}
+
+	repeat = 1;
+	if (args_has(args, 'R')) {
+		repeat = args_strtonum(args, 'R', 1, SHRT_MAX, &cause);
+		if (cause != NULL) {
+			cmdq_error(cmdq, "repeat count %s", cause);
+			free(cause);
+			return (CMD_RETURN_ERROR);
+		}
 	}
 
 	mtmp.key = key;
@@ -126,6 +164,7 @@ cmd_bind_key_mode_table(struct cmd *self, struct cmd_q *cmdq, int key)
 		mbind->mode = mtmp.mode;
 		RB_INSERT(mode_key_tree, mtab->tree, mbind);
 	}
+	mbind->repeat = repeat;
 	mbind->cmd = cmd;
 	mbind->arg = arg != NULL ? xstrdup(arg) : NULL;
 	return (CMD_RETURN_NORMAL);
