@@ -36,19 +36,28 @@ void	cmd_if_shell_done(struct cmd_q *);
 void	cmd_if_shell_free(void *);
 
 const struct cmd_entry cmd_if_shell_entry = {
-	"if-shell", "if",
-	"bFt:", 2, 3,
-	"[-bF] " CMD_TARGET_PANE_USAGE " shell-command command [command]",
-	0,
-	cmd_if_shell_exec
+	.name = "if-shell",
+	.alias = "if",
+
+	.args = { "bFt:", 2, 3 },
+	.usage = "[-bF] " CMD_TARGET_PANE_USAGE " shell-command command "
+		 "[command]",
+
+	.tflag = CMD_PANE_CANFAIL,
+
+	.flags = 0,
+	.exec = cmd_if_shell_exec
 };
 
 struct cmd_if_shell_data {
-	char		*cmd_if;
-	char		*cmd_else;
-	struct cmd_q	*cmdq;
-	int		 bflag;
-	int		 started;
+	char			*cmd_if;
+	char			*cmd_else;
+
+	struct cmd_q		*cmdq;
+	struct mouse_event	 mouse;
+
+	int			 bflag;
+	int			 references;
 };
 
 enum cmd_retval
@@ -58,25 +67,21 @@ cmd_if_shell_exec(struct cmd *self, struct cmd_q *cmdq)
 	struct cmd_if_shell_data	*cdata;
 	char				*shellcmd, *cmd, *cause;
 	struct cmd_list			*cmdlist;
-	struct client			*c;
-	struct session			*s = NULL;
-	struct winlink			*wl = NULL;
-	struct window_pane		*wp = NULL;
+	struct session			*s = cmdq->state.tflag.s;
+	struct winlink			*wl = cmdq->state.tflag.wl;
+	struct window_pane		*wp = cmdq->state.tflag.wp;
 	struct format_tree		*ft;
+	const char			*cwd;
 
-	if (args_has(args, 't'))
-		wl = cmd_find_pane(cmdq, args_get(args, 't'), &s, &wp);
-	else {
-		c = cmd_find_client(cmdq, NULL, 1);
-		if (c != NULL && c->session != NULL) {
-			s = c->session;
-			wl = s->curw;
-			wp = wl->window->active;
-		}
-	}
+	if (cmdq->client != NULL && cmdq->client->session == NULL)
+		cwd = cmdq->client->cwd;
+	else if (s != NULL)
+		cwd = s->cwd;
+	else
+		cwd = NULL;
 
-	ft = format_create();
-	format_defaults(ft, NULL, s, wl, wp);
+	ft = format_create(cmdq, 0);
+	format_defaults(ft, cmdq->state.c, s, wl, wp);
 	shellcmd = format_expand(ft, args->argv[0]);
 	format_free(ft);
 
@@ -86,6 +91,7 @@ cmd_if_shell_exec(struct cmd *self, struct cmd_q *cmdq)
 			cmd = args->argv[1];
 		else if (args->argc == 3)
 			cmd = args->argv[2];
+		free(shellcmd);
 		if (cmd == NULL)
 			return (CMD_RETURN_NORMAL);
 		if (cmd_string_parse(cmd, &cmdlist, NULL, 0, &cause) != 0) {
@@ -95,24 +101,28 @@ cmd_if_shell_exec(struct cmd *self, struct cmd_q *cmdq)
 			}
 			return (CMD_RETURN_ERROR);
 		}
-		cmdq_run(cmdq, cmdlist);
+		cmdq_run(cmdq, cmdlist, &cmdq->item->mouse);
 		cmd_list_free(cmdlist);
 		return (CMD_RETURN_NORMAL);
 	}
 
 	cdata = xmalloc(sizeof *cdata);
+
 	cdata->cmd_if = xstrdup(args->argv[1]);
 	if (args->argc == 3)
 		cdata->cmd_else = xstrdup(args->argv[2]);
 	else
 		cdata->cmd_else = NULL;
+
 	cdata->bflag = args_has(args, 'b');
 
-	cdata->started = 0;
 	cdata->cmdq = cmdq;
+	memcpy(&cdata->mouse, &cmdq->item->mouse, sizeof cdata->mouse);
 	cmdq->references++;
 
-	job_run(shellcmd, s, cmd_if_shell_callback, cmd_if_shell_free, cdata);
+	cdata->references = 1;
+	job_run(shellcmd, s, cwd, cmd_if_shell_callback, cmd_if_shell_free,
+	    cdata);
 	free(shellcmd);
 
 	if (cdata->bflag)
@@ -128,7 +138,7 @@ cmd_if_shell_callback(struct job *job)
 	struct cmd_list			*cmdlist;
 	char				*cause, *cmd;
 
-	if (cmdq->dead)
+	if (cmdq->flags & CMD_Q_DEAD)
 		return;
 
 	if (!WIFEXITED(job->status) || WEXITSTATUS(job->status) != 0)
@@ -146,13 +156,13 @@ cmd_if_shell_callback(struct job *job)
 		return;
 	}
 
-	cdata->started = 1;
-
 	cmdq1 = cmdq_new(cmdq->client);
+	cmdq1->flags |= cmdq->flags & CMD_Q_NOHOOKS;
 	cmdq1->emptyfn = cmd_if_shell_done;
 	cmdq1->data = cdata;
 
-	cmdq_run(cmdq1, cmdlist);
+	cdata->references++;
+	cmdq_run(cmdq1, cmdlist, &cdata->mouse);
 	cmd_list_free(cmdlist);
 }
 
@@ -164,12 +174,14 @@ cmd_if_shell_done(struct cmd_q *cmdq1)
 
 	if (cmdq1->client_exit >= 0)
 		cmdq->client_exit = cmdq1->client_exit;
+	cmdq_free(cmdq1);
+
+	if (--cdata->references != 0)
+		return;
 
 	if (!cmdq_free(cmdq) && !cdata->bflag)
 		cmdq_continue(cmdq);
 
-	cmdq_free(cmdq1);
-
 	free(cdata->cmd_else);
 	free(cdata->cmd_if);
 	free(cdata);
@@ -181,7 +193,7 @@ cmd_if_shell_free(void *data)
 	struct cmd_if_shell_data	*cdata = data;
 	struct cmd_q			*cmdq = cdata->cmdq;
 
-	if (cdata->started)
+	if (--cdata->references != 0)
 		return;
 
 	if (!cmdq_free(cmdq) && !cdata->bflag)
