@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -65,33 +65,42 @@ struct options_entry *cmd_set_option_style(struct cmd *, struct cmd_q *,
 	    const char *);
 
 const struct cmd_entry cmd_set_option_entry = {
-	"set-option", "set",
-	"agoqst:uw", 1, 2,
-	"[-agosquw] [-t target-session|target-window] option [value]",
-	0,
-	cmd_set_option_exec
+	.name = "set-option",
+	.alias = "set",
+
+	.args = { "agoqst:uw", 1, 2 },
+	.usage = "[-agosquw] [-t target-window] option [value]",
+
+	.tflag = CMD_WINDOW_CANFAIL,
+
+	.flags = 0,
+	.exec = cmd_set_option_exec
 };
 
 const struct cmd_entry cmd_set_window_option_entry = {
-	"set-window-option", "setw",
-	"agoqt:u", 1, 2,
-	"[-agoqu] " CMD_TARGET_WINDOW_USAGE " option [value]",
-	0,
-	cmd_set_option_exec
+	.name = "set-window-option",
+	.alias = "setw",
+
+	.args = { "agoqt:u", 1, 2 },
+	.usage = "[-agoqu] " CMD_TARGET_WINDOW_USAGE " option [value]",
+
+	.tflag = CMD_WINDOW_CANFAIL,
+
+	.flags = 0,
+	.exec = cmd_set_option_exec
 };
 
 enum cmd_retval
 cmd_set_option_exec(struct cmd *self, struct cmd_q *cmdq)
 {
 	struct args				*args = self->args;
-	const struct options_table_entry	*table, *oe;
-	struct session				*s;
-	struct winlink				*wl;
-	struct client				*c;
-	struct options				*oo;
+	struct session				*s = cmdq->state.tflag.s;
+	struct winlink				*wl = cmdq->state.tflag.wl;
 	struct window				*w;
-	const char				*optstr, *valstr;
-	u_int					 i;
+	struct client				*c;
+	const struct options_table_entry	*oe;
+	struct options				*oo;
+	const char				*optstr, *valstr, *target;
 
 	/* Get the option name and value. */
 	optstr = args->argv[0];
@@ -109,10 +118,13 @@ cmd_set_option_exec(struct cmd *self, struct cmd_q *cmdq)
 		return (cmd_set_option_user(self, cmdq, optstr, valstr));
 
 	/* Find the option entry, try each table. */
-	table = oe = NULL;
-	if (options_table_find(optstr, &table, &oe) != 0) {
-		cmdq_error(cmdq, "ambiguous option: %s", optstr);
-		return (CMD_RETURN_ERROR);
+	oe = NULL;
+	if (options_table_find(optstr, &oe) != 0) {
+		if (!args_has(args, 'q')) {
+			cmdq_error(cmdq, "ambiguous option: %s", optstr);
+			return (CMD_RETURN_ERROR);
+		}
+		return (CMD_RETURN_NORMAL);
 	}
 	if (oe == NULL) {
 		if (!args_has(args, 'q')) {
@@ -122,37 +134,35 @@ cmd_set_option_exec(struct cmd *self, struct cmd_q *cmdq)
 		return (CMD_RETURN_NORMAL);
 	}
 
-	/* Work out the tree from the table. */
-	if (table == server_options_table)
-		oo = &global_options;
-	else if (table == window_options_table) {
+	/* Work out the tree from the scope of the option. */
+	if (oe->scope == OPTIONS_TABLE_SERVER)
+		oo = global_options;
+	else if (oe->scope == OPTIONS_TABLE_WINDOW) {
 		if (args_has(self->args, 'g'))
-			oo = &global_w_options;
-		else {
-			wl = cmd_find_window(cmdq, args_get(args, 't'), NULL);
-			if (wl == NULL) {
-				cmdq_error(cmdq,
-				    "couldn't set '%s'%s", optstr,
-				    (!args_has(args, 't') && !args_has(args,
-				    'g')) ? " need target window or -g" : "");
-				return (CMD_RETURN_ERROR);
-			}
-			oo = &wl->window->options;
-		}
-	} else if (table == session_options_table) {
+			oo = global_w_options;
+		else if (wl == NULL) {
+			target = args_get(args, 't');
+			if (target != NULL) {
+				cmdq_error(cmdq, "no such window: %s",
+				    target);
+			} else
+				cmdq_error(cmdq, "no current window");
+			return (CMD_RETURN_ERROR);
+		} else
+			oo = wl->window->options;
+	} else if (oe->scope == OPTIONS_TABLE_SESSION) {
 		if (args_has(self->args, 'g'))
-			oo = &global_s_options;
-		else {
-			s = cmd_find_session(cmdq, args_get(args, 't'), 0);
-			if (s == NULL) {
-				cmdq_error(cmdq,
-				    "couldn't set '%s'%s", optstr,
-				    (!args_has(args, 't') && !args_has(args,
-				    'g')) ? " need target session or -g" : "");
-				return (CMD_RETURN_ERROR);
-			}
-			oo = &s->options;
-		}
+			oo = global_s_options;
+		else if (s == NULL) {
+			target = args_get(args, 't');
+			if (target != NULL) {
+				cmdq_error(cmdq, "no such session: %s",
+				    target);
+			} else
+				cmdq_error(cmdq, "no current session");
+			return (CMD_RETURN_ERROR);
+		} else
+			oo = s->options;
 	} else {
 		cmdq_error(cmdq, "unknown table");
 		return (CMD_RETURN_ERROR);
@@ -174,23 +184,38 @@ cmd_set_option_exec(struct cmd *self, struct cmd_q *cmdq)
 			return (CMD_RETURN_ERROR);
 	}
 
-	/* Start or stop timers when automatic-rename changed. */
+	/* Start or stop timers if necessary. */
 	if (strcmp(oe->name, "automatic-rename") == 0) {
-		for (i = 0; i < ARRAY_LENGTH(&windows); i++) {
-			if ((w = ARRAY_ITEM(&windows, i)) == NULL)
-				continue;
-			if (options_get_number(&w->options, "automatic-rename"))
-				queue_window_name(w);
-			else if (event_initialized(&w->name_timer))
-				evtimer_del(&w->name_timer);
+		RB_FOREACH(w, windows, &windows) {
+			if (options_get_number(w->options, "automatic-rename"))
+				w->active->flags |= PANE_CHANGED;
 		}
 	}
+	if (strcmp(oe->name, "key-table") == 0) {
+		TAILQ_FOREACH(c, &clients, entry)
+			server_client_set_key_table(c, NULL);
+	}
+	if (strcmp(oe->name, "status") == 0 ||
+	    strcmp(oe->name, "status-interval") == 0)
+		status_timer_start_all();
+	if (strcmp(oe->name, "monitor-silence") == 0)
+		alerts_reset_all();
+	if (strcmp(oe->name, "window-style") == 0 ||
+	    strcmp(oe->name, "window-active-style") == 0) {
+		RB_FOREACH(w, windows, &windows)
+			w->flags |= WINDOW_STYLECHANGED;
+	}
+
+	/* When the pane-border-status option has been changed, resize panes. */
+	if (strcmp(oe->name, "pane-border-status") == 0) {
+		RB_FOREACH(w, windows, &windows)
+			layout_fix_panes(w, w->sx, w->sy);
+	}
 
 	/* Update sizes and redraw. May not need it but meh. */
 	recalculate_sizes();
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c != NULL && c->session != NULL)
+	TAILQ_FOREACH(c, &clients, entry) {
+		if (c->session != NULL)
 			server_redraw_client(c);
 	}
 
@@ -202,32 +227,25 @@ enum cmd_retval
 cmd_set_option_user(struct cmd *self, struct cmd_q *cmdq, const char *optstr,
     const char *valstr)
 {
-	struct args	*args = self->args;
-	struct session	*s;
-	struct winlink	*wl;
-	struct options	*oo;
+	struct args		*args = self->args;
+	struct session		*s = cmdq->state.tflag.s;
+	struct winlink		*wl = cmdq->state.tflag.wl;
+	struct options		*oo;
+	struct options_entry	*o;
 
 	if (args_has(args, 's'))
-		oo = &global_options;
+		oo = global_options;
 	else if (args_has(self->args, 'w') ||
 	    self->entry == &cmd_set_window_option_entry) {
 		if (args_has(self->args, 'g'))
-			oo = &global_w_options;
-		else {
-			wl = cmd_find_window(cmdq, args_get(args, 't'), NULL);
-			if (wl == NULL)
-				return (CMD_RETURN_ERROR);
-			oo = &wl->window->options;
-		}
+			oo = global_w_options;
+		else
+			oo = wl->window->options;
 	} else {
 		if (args_has(self->args, 'g'))
-			oo = &global_s_options;
-		else {
-			s = cmd_find_session(cmdq, args_get(args, 't'), 0);
-			if (s == NULL)
-				return (CMD_RETURN_ERROR);
-			oo = &s->options;
-		}
+			oo = global_s_options;
+		else
+			oo = s->options;
 	}
 
 	if (args_has(args, 'u')) {
@@ -245,23 +263,26 @@ cmd_set_option_user(struct cmd *self, struct cmd_q *cmdq, const char *optstr,
 		}
 		options_remove(oo, optstr);
 	} else {
-		if (valstr == NULL) {
-			cmdq_error(cmdq, "empty value");
-			return (CMD_RETURN_ERROR);
-		}
-		if (args_has(args, 'o') && options_find1(oo, optstr) != NULL) {
+		o = options_find1(oo, optstr);
+		if (args_has(args, 'o') && o != NULL) {
 			if (!args_has(args, 'q')) {
 				cmdq_error(cmdq, "already set: %s", optstr);
 				return (CMD_RETURN_ERROR);
 			}
 			return (CMD_RETURN_NORMAL);
 		}
-		options_set_string(oo, optstr, "%s", valstr);
+		if (valstr == NULL) {
+			cmdq_error(cmdq, "empty value");
+			return (CMD_RETURN_ERROR);
+		}
+		if (o != NULL && args_has(args, 'a'))
+			options_set_string(oo, optstr, "%s%s", o->str, valstr);
+		else
+			options_set_string(oo, optstr, "%s", valstr);
 	}
 	return (CMD_RETURN_NORMAL);
 }
 
-
 /* Unset an option. */
 int
 cmd_set_option_unset(struct cmd *self, struct cmd_q *cmdq,
@@ -270,16 +291,25 @@ cmd_set_option_unset(struct cmd *self, struct cmd_q *cmdq,
 {
 	struct args	*args = self->args;
 
-	if (args_has(args, 'g')) {
-		cmdq_error(cmdq, "can't unset global option: %s", oe->name);
-		return (-1);
-	}
 	if (value != NULL) {
 		cmdq_error(cmdq, "value passed to unset option: %s", oe->name);
 		return (-1);
 	}
 
-	options_remove(oo, oe->name);
+	if (args_has(args, 'g') || oo == global_options) {
+		switch (oe->type) {
+		case OPTIONS_TABLE_STRING:
+			options_set_string(oo, oe->name, "%s", oe->default_str);
+			break;
+		case OPTIONS_TABLE_STYLE:
+			options_set_style(oo, oe->name, oe->default_str, 0);
+			break;
+		default:
+			options_set_number(oo, oe->name, oe->default_num);
+			break;
+		}
+	} else
+		options_remove(oo, oe->name);
 	return (0);
 }
 
@@ -291,9 +321,15 @@ cmd_set_option_set(struct cmd *self, struct cmd_q *cmdq,
 {
 	struct options_entry	*o;
 
-	if (oe->type != OPTIONS_TABLE_FLAG && value == NULL) {
-		cmdq_error(cmdq, "empty value");
-		return (-1);
+	switch (oe->type) {
+	case OPTIONS_TABLE_FLAG:
+	case OPTIONS_TABLE_CHOICE:
+		break;
+	default:
+		if (value == NULL) {
+			cmdq_error(cmdq, "empty value");
+			return (-1);
+		}
 	}
 
 	o = NULL;
@@ -334,7 +370,7 @@ cmd_set_option_set(struct cmd *self, struct cmd_q *cmdq,
 
 /* Set a string option. */
 struct options_entry *
-cmd_set_option_string(struct cmd *self, unused struct cmd_q *cmdq,
+cmd_set_option_string(struct cmd *self, __unused struct cmd_q *cmdq,
     const struct options_table_entry *oe, struct options *oo,
     const char *value)
 {
@@ -356,7 +392,7 @@ cmd_set_option_string(struct cmd *self, unused struct cmd_q *cmdq,
 
 /* Set a number option. */
 struct options_entry *
-cmd_set_option_number(unused struct cmd *self, struct cmd_q *cmdq,
+cmd_set_option_number(__unused struct cmd *self, struct cmd_q *cmdq,
     const struct options_table_entry *oe, struct options *oo,
     const char *value)
 {
@@ -374,13 +410,14 @@ cmd_set_option_number(unused struct cmd *self, struct cmd_q *cmdq,
 
 /* Set a key option. */
 struct options_entry *
-cmd_set_option_key(unused struct cmd *self, struct cmd_q *cmdq,
+cmd_set_option_key(__unused struct cmd *self, struct cmd_q *cmdq,
     const struct options_table_entry *oe, struct options *oo,
     const char *value)
 {
-	int	key;
+	key_code	key;
 
-	if ((key = key_string_lookup_string(value)) == KEYC_NONE) {
+	key = key_string_lookup_string(value);
+	if (key == KEYC_UNKNOWN) {
 		cmdq_error(cmdq, "bad key: %s", value);
 		return (NULL);
 	}
@@ -390,7 +427,7 @@ cmd_set_option_key(unused struct cmd *self, struct cmd_q *cmdq,
 
 /* Set a colour option. */
 struct options_entry *
-cmd_set_option_colour(unused struct cmd *self, struct cmd_q *cmdq,
+cmd_set_option_colour(__unused struct cmd *self, struct cmd_q *cmdq,
     const struct options_table_entry *oe, struct options *oo,
     const char *value)
 {
@@ -406,7 +443,7 @@ cmd_set_option_colour(unused struct cmd *self, struct cmd_q *cmdq,
 
 /* Set an attributes option. */
 struct options_entry *
-cmd_set_option_attributes(unused struct cmd *self, struct cmd_q *cmdq,
+cmd_set_option_attributes(__unused struct cmd *self, struct cmd_q *cmdq,
     const struct options_table_entry *oe, struct options *oo,
     const char *value)
 {
@@ -422,7 +459,7 @@ cmd_set_option_attributes(unused struct cmd *self, struct cmd_q *cmdq,
 
 /* Set a flag option. */
 struct options_entry *
-cmd_set_option_flag(unused struct cmd *self, struct cmd_q *cmdq,
+cmd_set_option_flag(__unused struct cmd *self, struct cmd_q *cmdq,
     const struct options_table_entry *oe, struct options *oo,
     const char *value)
 {
@@ -450,28 +487,34 @@ cmd_set_option_flag(unused struct cmd *self, struct cmd_q *cmdq,
 
 /* Set a choice option. */
 struct options_entry *
-cmd_set_option_choice(unused struct cmd *self, struct cmd_q *cmdq,
+cmd_set_option_choice(__unused struct cmd *self, struct cmd_q *cmdq,
     const struct options_table_entry *oe, struct options *oo,
     const char *value)
 {
 	const char	**choicep;
 	int		  n, choice = -1;
 
-	n = 0;
-	for (choicep = oe->choices; *choicep != NULL; choicep++) {
-		n++;
-		if (strncmp(*choicep, value, strlen(value)) != 0)
-			continue;
+	if (value == NULL) {
+		choice = options_get_number(oo, oe->name);
+		if (choice < 2)
+			choice = !choice;
+	} else {
+		n = 0;
+		for (choicep = oe->choices; *choicep != NULL; choicep++) {
+			n++;
+			if (strncmp(*choicep, value, strlen(value)) != 0)
+				continue;
 
-		if (choice != -1) {
-			cmdq_error(cmdq, "ambiguous value: %s", value);
+			if (choice != -1) {
+				cmdq_error(cmdq, "ambiguous value: %s", value);
+				return (NULL);
+			}
+			choice = n - 1;
+		}
+		if (choice == -1) {
+			cmdq_error(cmdq, "unknown value: %s", value);
 			return (NULL);
 		}
-		choice = n - 1;
-	}
-	if (choice == -1) {
-		cmdq_error(cmdq, "unknown value: %s", value);
-		return (NULL);
 	}
 
 	return (options_set_number(oo, oe->name, choice));
