@@ -1,4 +1,4 @@
-/* $Id$ */
+/* $OpenBSD$ */
 
 /*
  * Copyright (c) 2008 Tiago Cunha <me@tiagocunha.org>
@@ -28,16 +28,17 @@
 
 enum cmd_retval	cmd_source_file_exec(struct cmd *, struct cmd_q *);
 
-void		cmd_source_file_show(struct cmd_q *);
 void		cmd_source_file_done(struct cmd_q *);
 
 const struct cmd_entry cmd_source_file_entry = {
-	"source-file", "source",
-	"", 1, 1,
-	"path",
-	0,
-	NULL,
-	cmd_source_file_exec
+	.name = "source-file",
+	.alias = "source",
+
+	.args = { "q", 1, 1 },
+	.usage = "[-q] path",
+
+	.flags = 0,
+	.exec = cmd_source_file_exec
 };
 
 enum cmd_retval
@@ -45,27 +46,26 @@ cmd_source_file_exec(struct cmd *self, struct cmd_q *cmdq)
 {
 	struct args	*args = self->args;
 	struct cmd_q	*cmdq1;
-	char		*cause;
+	int		 quiet;
 
-	cmdq1 = cmdq_new(NULL);
-	cmdq1->client = cmdq->client;
+	cmdq1 = cmdq_new(cmdq->client);
+	cmdq1->flags |= cmdq->flags & CMD_Q_NOHOOKS;
 	cmdq1->emptyfn = cmd_source_file_done;
 	cmdq1->data = cmdq;
 
-	switch (load_cfg(args->argv[0], cmdq1, &cause)) {
+	quiet = args_has(args, 'q');
+	switch (load_cfg(args->argv[0], cmdq1, quiet)) {
 	case -1:
+		cmdq_free(cmdq1);
 		if (cfg_references == 0) {
-			cmdq_free(cmdq1);
-			cmdq_error(cmdq, "%s", cause);
-			free(cause);
+			cfg_print_causes(cmdq);
 			return (CMD_RETURN_ERROR);
 		}
-		ARRAY_ADD(&cfg_causes, cause);
-		/* FALLTHROUGH */
+		return (CMD_RETURN_NORMAL);
 	case 0:
-		if (cfg_references == 0)
-			cmd_source_file_show(cmdq);
 		cmdq_free(cmdq1);
+		if (cfg_references == 0)
+			cfg_print_causes(cmdq);
 		return (CMD_RETURN_NORMAL);
 	}
 
@@ -76,20 +76,6 @@ cmd_source_file_exec(struct cmd *self, struct cmd_q *cmdq)
 	return (CMD_RETURN_WAIT);
 }
 
-void
-cmd_source_file_show(struct cmd_q *cmdq)
-{
-	u_int	 i;
-	char	*cause;
-
-	for (i = 0; i < ARRAY_LENGTH(&cfg_causes); i++) {
-		cause = ARRAY_ITEM(&cfg_causes, i);
-		cmdq_print(cmdq, "%s", cause);
-		free(cause);
-	}
-	ARRAY_FREE(&cfg_causes);
-}
-
 void
 cmd_source_file_done(struct cmd_q *cmdq1)
 {
@@ -106,6 +92,6 @@ cmd_source_file_done(struct cmd_q *cmdq1)
 		return;
 
 	if (cfg_references == 0)
-		cmd_source_file_show(cmdq);
+		cfg_print_causes(cmdq);
 	cmdq_continue(cmdq);
 }
