@@ -52,7 +52,7 @@ char	*BSDoptarg;		/* argument associated with option */
 int
 BSDgetopt(int nargc, char *const *nargv, const char *ostr)
 {
-	static char *place = EMSG;		/* option letter processing */
+	static const char *place = EMSG;	/* option letter processing */
 	char *oli;				/* option letter list index */
 
 	if (ostr == NULL)
@@ -84,7 +84,8 @@ BSDgetopt(int nargc, char *const *nargv, const char *ostr)
 			++BSDoptind;
 		if (BSDopterr && *ostr != ':')
 			(void)fprintf(stderr,
-			    "%s: unknown option -- %c\n", __progname, BSDoptopt);
+			    "%s: unknown option -- %c\n", getprogname(),
+			    BSDoptopt);
 		return (BADCH);
 	}
 	if (*++oli != ':') {			/* don't need argument */
@@ -94,7 +95,7 @@ BSDgetopt(int nargc, char *const *nargv, const char *ostr)
 	}
 	else {					/* need an argument */
 		if (*place)			/* no white space */
-			BSDoptarg = place;
+			BSDoptarg = (char *)place;
 		else if (nargc <= ++BSDoptind) {	/* no arg */
 			place = EMSG;
 			if (*ostr == ':')
@@ -102,10 +103,10 @@ BSDgetopt(int nargc, char *const *nargv, const char *ostr)
 			if (BSDopterr)
 				(void)fprintf(stderr,
 				    "%s: option requires an argument -- %c\n",
-				    __progname, BSDoptopt);
+				    getprogname(), BSDoptopt);
 			return (BADCH);
 		}
-	 	else				/* white space */
+		else				/* white space */
 			BSDoptarg = nargv[BSDoptind];
 		place = EMSG;
 		++BSDoptind;
