@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2009 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2009 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -27,6 +27,9 @@
  * Environment - manipulate a set of environment variables.
  */
 
+RB_HEAD(environ, environ_entry);
+int	environ_cmp(struct environ_entry *, struct environ_entry *);
+RB_PROTOTYPE(environ, environ_entry, entry, environ_cmp);
 RB_GENERATE(environ, environ_entry, entry, environ_cmp);
 
 int
@@ -36,25 +39,42 @@ environ_cmp(struct environ_entry *envent1, struct environ_entry *envent2)
 }
 
 /* Initialise the environment. */
-void
-environ_init(struct environ *env)
+struct environ *
+environ_create(void)
 {
+	struct environ	*env;
+
+	env = xcalloc(1, sizeof *env);
 	RB_INIT(env);
+
+	return (env);
 }
 
 /* Free an environment. */
 void
 environ_free(struct environ *env)
 {
-	struct environ_entry	*envent;
+	struct environ_entry	*envent, *envent1;
 
-	while (!RB_EMPTY(env)) {
-		envent = RB_ROOT(env);
+	RB_FOREACH_SAFE(envent, environ, env, envent1) {
 		RB_REMOVE(environ, env, envent);
 		free(envent->name);
 		free(envent->value);
 		free(envent);
 	}
+	free(env);
+}
+
+struct environ_entry *
+environ_first(struct environ *env)
+{
+	return (RB_MIN(environ, env));
+}
+
+struct environ_entry *
+environ_next(struct environ_entry *envent)
+{
+	return (RB_NEXT(environ, env, envent));
 }
 
 /* Copy one environment into another. */
@@ -63,8 +83,12 @@ environ_copy(struct environ *srcenv, struct environ *dstenv)
 {
 	struct environ_entry	*envent;
 
-	RB_FOREACH(envent, environ, srcenv)
-		environ_set(dstenv, envent->name, envent->value);
+	RB_FOREACH(envent, environ, srcenv) {
+		if (envent->value == NULL)
+			environ_clear(dstenv, envent->name);
+		else
+			environ_set(dstenv, envent->name, "%s", envent->value);
+	}
 }
 
 /* Find an environment variable. */
@@ -79,23 +103,37 @@ environ_find(struct environ *env, const char *name)
 
 /* Set an environment variable. */
 void
-environ_set(struct environ *env, const char *name, const char *value)
+environ_set(struct environ *env, const char *name, const char *fmt, ...)
+{
+	struct environ_entry	*envent;
+	va_list			 ap;
+
+	va_start(ap, fmt);
+	if ((envent = environ_find(env, name)) != NULL) {
+		free(envent->value);
+		xvasprintf(&envent->value, fmt, ap);
+	} else {
+		envent = xmalloc(sizeof *envent);
+		envent->name = xstrdup(name);
+		xvasprintf(&envent->value, fmt, ap);
+		RB_INSERT(environ, env, envent);
+	}
+	va_end(ap);
+}
+
+/* Clear an environment variable. */
+void
+environ_clear(struct environ *env, const char *name)
 {
 	struct environ_entry	*envent;
 
 	if ((envent = environ_find(env, name)) != NULL) {
 		free(envent->value);
-		if (value != NULL)
-			envent->value = xstrdup(value);
-		else
-			envent->value = NULL;
+		envent->value = NULL;
 	} else {
 		envent = xmalloc(sizeof *envent);
 		envent->name = xstrdup(name);
-		if (value != NULL)
-			envent->value = xstrdup(value);
-		else
-			envent->value = NULL;
+		envent->value = NULL;
 		RB_INSERT(environ, env, envent);
 	}
 }
@@ -114,7 +152,7 @@ environ_put(struct environ *env, const char *var)
 	name = xstrdup(var);
 	name[strcspn(name, "=")] = '\0';
 
-	environ_set(env, name, value);
+	environ_set(env, name, "%s", value);
 	free(name);
 }
 
@@ -146,9 +184,9 @@ environ_update(const char *vars, struct environ *srcenv,
 	copyvars = next = xstrdup(vars);
 	while ((var = strsep(&next, " ")) != NULL) {
 		if ((envent = environ_find(srcenv, var)) == NULL)
-			environ_set(dstenv, var, NULL);
+			environ_clear(dstenv, var);
 		else
-			environ_set(dstenv, envent->name, envent->value);
+			environ_set(dstenv, envent->name, "%s", envent->value);
 	}
 	free(copyvars);
 }
@@ -157,19 +195,25 @@ environ_update(const char *vars, struct environ *srcenv,
 void
 environ_push(struct environ *env)
 {
-	struct environ_entry	 *envent;
-	char			**vp, *v;
-
-	for (vp = environ; *vp != NULL; vp++) {
-		v = xstrdup(*vp);
-		v[strcspn(v, "=")] = '\0';
-
-		unsetenv(v);
-		free(v);
-	}
+	struct environ_entry	*envent;
 
+	environ = xcalloc(1, sizeof *environ);
 	RB_FOREACH(envent, environ, env) {
-		if (envent->value != NULL)
+		if (envent->value != NULL && *envent->name != '\0')
 			setenv(envent->name, envent->value, 1);
 	}
 }
+
+/* Log the environment. */
+void
+environ_log(struct environ *env, const char *prefix)
+{
+	struct environ_entry	*envent;
+
+	RB_FOREACH(envent, environ, env) {
+		if (envent->value != NULL && *envent->name != '\0') {
+			log_debug("%s%s=%s", prefix, envent->name,
+			    envent->value);
+		}
+	}
+}
