@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -22,12 +22,12 @@
 
 #include "tmux.h"
 
-int	key_string_search_table(const char *);
-int	key_string_get_modifiers(const char **);
+static key_code	key_string_search_table(const char *);
+static key_code	key_string_get_modifiers(const char **);
 
 const struct {
 	const char     *string;
-	int	 	key;
+	key_code	key;
 } key_string_table[] = {
 	/* Function keys. */
 	{ "F1",		KEYC_F1 },
@@ -82,10 +82,26 @@ const struct {
 	{ "KPEnter",	KEYC_KP_ENTER },
 	{ "KP0",	KEYC_KP_ZERO },
 	{ "KP.",	KEYC_KP_PERIOD },
+
+	/* Mouse keys. */
+	KEYC_MOUSE_STRING(MOUSEDOWN1, MouseDown1),
+	KEYC_MOUSE_STRING(MOUSEDOWN2, MouseDown2),
+	KEYC_MOUSE_STRING(MOUSEDOWN3, MouseDown3),
+	KEYC_MOUSE_STRING(MOUSEUP1, MouseUp1),
+	KEYC_MOUSE_STRING(MOUSEUP2, MouseUp2),
+	KEYC_MOUSE_STRING(MOUSEUP3, MouseUp3),
+	KEYC_MOUSE_STRING(MOUSEDRAG1, MouseDrag1),
+	KEYC_MOUSE_STRING(MOUSEDRAG2, MouseDrag2),
+	KEYC_MOUSE_STRING(MOUSEDRAG3, MouseDrag3),
+	KEYC_MOUSE_STRING(MOUSEDRAGEND1, MouseDragEnd1),
+	KEYC_MOUSE_STRING(MOUSEDRAGEND2, MouseDragEnd2),
+	KEYC_MOUSE_STRING(MOUSEDRAGEND3, MouseDragEnd3),
+	KEYC_MOUSE_STRING(WHEELUP, WheelUp),
+	KEYC_MOUSE_STRING(WHEELDOWN, WheelDown),
 };
 
 /* Find key string in table. */
-int
+static key_code
 key_string_search_table(const char *string)
 {
 	u_int	i;
@@ -94,14 +110,14 @@ key_string_search_table(const char *string)
 		if (strcasecmp(string, key_string_table[i].string) == 0)
 			return (key_string_table[i].key);
 	}
-	return (KEYC_NONE);
+	return (KEYC_UNKNOWN);
 }
 
 /* Find modifiers. */
-int
+static key_code
 key_string_get_modifiers(const char **string)
 {
-	int	modifiers;
+	key_code	modifiers;
 
 	modifiers = 0;
 	while (((*string)[0] != '\0') && (*string)[1] == '-') {
@@ -125,18 +141,28 @@ key_string_get_modifiers(const char **string)
 }
 
 /* Lookup a string and convert to a key value. */
-int
+key_code
 key_string_lookup_string(const char *string)
 {
 	static const char	*other = "!#()+,-.0123456789:;<=>?'\r\t";
-	int			 key, modifiers;
-	u_short			 u;
-	int			 size;
+	key_code		 key;
+	u_int			 u;
+	key_code		 modifiers;
+	struct utf8_data	 ud;
+	u_int			 i;
+	enum utf8_state		 more;
+	wchar_t			 wc;
+
+	/* Is this no key? */
+	if (strcasecmp(string, "None") == 0)
+		return (KEYC_NONE);
 
 	/* Is this a hexadecimal value? */
 	if (string[0] == '0' && string[1] == 'x') {
-	        if (sscanf(string + 2, "%hx%n", &u, &size) != 1 || size > 4)
-	                return (KEYC_NONE);
+	        if (sscanf(string + 2, "%x", &u) != 1)
+	                return (KEYC_UNKNOWN);
+		if (u > 0x1fffff)
+	                return (KEYC_UNKNOWN);
 	        return (u);
 	}
 
@@ -148,18 +174,31 @@ key_string_lookup_string(const char *string)
 	}
 	modifiers |= key_string_get_modifiers(&string);
 	if (string[0] == '\0')
-		return (KEYC_NONE);
+		return (KEYC_UNKNOWN);
 
 	/* Is this a standard ASCII key? */
-	if (string[1] == '\0') {
-		key = (u_char) string[0];
-		if (key < 32 || key == 127 || key > 255)
-			return (KEYC_NONE);
+	if (string[1] == '\0' && (u_char)string[0] <= 127) {
+		key = (u_char)string[0];
+		if (key < 32 || key == 127)
+			return (KEYC_UNKNOWN);
 	} else {
+		/* Try as a UTF-8 key. */
+		if ((more = utf8_open(&ud, (u_char)*string)) == UTF8_MORE) {
+			if (strlen(string) != ud.size)
+				return (KEYC_UNKNOWN);
+			for (i = 1; i < ud.size; i++)
+				more = utf8_append(&ud, (u_char)string[i]);
+			if (more != UTF8_DONE)
+				return (KEYC_UNKNOWN);
+			if (utf8_combine(&ud, &wc) != UTF8_DONE)
+				return (KEYC_UNKNOWN);
+			return (wc | modifiers);
+		}
+
 		/* Otherwise look the key up in the table. */
 		key = key_string_search_table(string);
-		if (key == KEYC_NONE)
-			return (KEYC_NONE);
+		if (key == KEYC_UNKNOWN)
+			return (KEYC_UNKNOWN);
 	}
 
 	/* Convert the standard control keys. */
@@ -173,7 +212,7 @@ key_string_lookup_string(const char *string)
 		else if (key == 63)
 			key = KEYC_BSPACE;
 		else
-			return (KEYC_NONE);
+			return (KEYC_UNKNOWN);
 		modifiers &= ~KEYC_CTRL;
 	}
 
@@ -182,17 +221,25 @@ key_string_lookup_string(const char *string)
 
 /* Convert a key code into string format, with prefix if necessary. */
 const char *
-key_string_lookup_key(int key)
+key_string_lookup_key(key_code key)
 {
-	static char	out[24];
-	char		tmp[8];
-	u_int		i;
+	static char		out[24];
+	char			tmp[8];
+	u_int			i;
+	struct utf8_data	ud;
+	size_t			off;
 
 	*out = '\0';
 
 	/* Handle no key. */
 	if (key == KEYC_NONE)
-		return ("none");
+		return ("None");
+
+	/* Handle special keys. */
+	if (key == KEYC_UNKNOWN)
+		return ("Unknown");
+	if (key == KEYC_MOUSE)
+		return ("Mouse");
 
 	/*
 	 * Special case: display C-@ as C-Space. Could do this below in
@@ -222,21 +269,33 @@ key_string_lookup_key(int key)
 		return (out);
 	}
 
+	/* Is this a UTF-8 key? */
+	if (key > 127 && key < KEYC_BASE) {
+		if (utf8_split(key, &ud) == UTF8_DONE) {
+			off = strlen(out);
+			memcpy(out + off, ud.data, ud.size);
+			out[off + ud.size] = '\0';
+			return (out);
+		}
+	}
+
 	/* Invalid keys are errors. */
-	if (key == 127 || key > 255)
-		return (NULL);
+	if (key == 127 || key > 255) {
+		snprintf(out, sizeof out, "Invalid#%llx", key);
+		return (out);
+	}
 
 	/* Check for standard or control key. */
-	if (key >= 0 && key <= 32) {
+	if (key <= 32) {
 		if (key == 0 || key > 26)
-			xsnprintf(tmp, sizeof tmp, "C-%c", 64 + key);
+			xsnprintf(tmp, sizeof tmp, "C-%c", (int)(64 + key));
 		else
-			xsnprintf(tmp, sizeof tmp, "C-%c", 96 + key);
+			xsnprintf(tmp, sizeof tmp, "C-%c", (int)(96 + key));
 	} else if (key >= 32 && key <= 126) {
 		tmp[0] = key;
 		tmp[1] = '\0';
 	} else if (key >= 128)
-		xsnprintf(tmp, sizeof tmp, "\\%o", key);
+		xsnprintf(tmp, sizeof tmp, "\\%llo", key);
 
 	strlcat(out, tmp, sizeof out);
 	return (out);
