@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2010 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2010 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -23,14 +23,17 @@
 
 #include "tmux.h"
 
-struct layout_cell     *layout_find_bottomright(struct layout_cell *);
-u_short			layout_checksum(const char *);
-int			layout_append(struct layout_cell *, char *, size_t);
-struct layout_cell     *layout_construct(struct layout_cell *, const char **);
-void			layout_assign(struct window_pane **, struct layout_cell *);
+static struct layout_cell	*layout_find_bottomright(struct layout_cell *);
+static u_short			 layout_checksum(const char *);
+static int			 layout_append(struct layout_cell *, char *,
+				     size_t);
+static struct layout_cell	*layout_construct(struct layout_cell *,
+				     const char **);
+static void			 layout_assign(struct window_pane **,
+				     struct layout_cell *);
 
 /* Find the bottom-right cell. */
-struct layout_cell *
+static struct layout_cell *
 layout_find_bottomright(struct layout_cell *lc)
 {
 	if (lc->type == LAYOUT_WINDOWPANE)
@@ -40,7 +43,7 @@ layout_find_bottomright(struct layout_cell *lc)
 }
 
 /* Calculate layout checksum. */
-u_short
+static u_short
 layout_checksum(const char *layout)
 {
 	u_short	csum;
@@ -55,20 +58,20 @@ layout_checksum(const char *layout)
 
 /* Dump layout as a string. */
 char *
-layout_dump(struct window *w)
+layout_dump(struct layout_cell *root)
 {
 	char	layout[BUFSIZ], *out;
 
 	*layout = '\0';
-	if (layout_append(w->layout_root, layout, sizeof layout) != 0)
+	if (layout_append(root, layout, sizeof layout) != 0)
 		return (NULL);
 
-	xasprintf(&out, "%04x,%s", layout_checksum(layout), layout);
+	xasprintf(&out, "%04hx,%s", layout_checksum(layout), layout);
 	return (out);
 }
 
 /* Append information for a single cell. */
-int
+static int
 layout_append(struct layout_cell *lc, char *buf, size_t len)
 {
 	struct layout_cell     *lcchild;
@@ -147,7 +150,7 @@ layout_parse(struct window *w, const char *layout)
 
 		/* Fewer panes than cells - close the bottom right. */
 		lcchild = layout_find_bottomright(lc);
-		layout_destroy_cell(lcchild, &lc);
+		layout_destroy_cell(w, lcchild, &lc);
 	}
 
 	/* Save the old window size and resize to the layout size. */
@@ -182,7 +185,7 @@ fail:
 }
 
 /* Assign panes into cells. */
-void
+static void
 layout_assign(struct window_pane **wp, struct layout_cell *lc)
 {
 	struct layout_cell	*lcchild;
@@ -201,7 +204,7 @@ layout_assign(struct window_pane **wp, struct layout_cell *lc)
 }
 
 /* Construct a cell from all or part of a layout tree. */
-struct layout_cell *
+static struct layout_cell *
 layout_construct(struct layout_cell *lcparent, const char **layout)
 {
 	struct layout_cell     *lc, *lcchild;
