@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -17,6 +17,7 @@
  */
 
 #include <sys/types.h>
+#include <sys/uio.h>
 
 #include <stdlib.h>
 #include <string.h>
@@ -31,61 +32,19 @@ void		server_callback_identify(int, short, void *);
 void
 server_fill_environ(struct session *s, struct environ *env)
 {
-	char	var[PATH_MAX], *term;
-	u_int	idx;
-	long	pid;
+	char	*term;
+	u_int	 idx;
+	long	 pid;
 
 	if (s != NULL) {
-		term = options_get_string(&s->options, "default-terminal");
-		environ_set(env, "TERM", term);
+		term = options_get_string(global_options, "default-terminal");
+		environ_set(env, "TERM", "%s", term);
 
 		idx = s->id;
 	} else
-		idx = -1;
+		idx = (u_int)-1;
 	pid = getpid();
-	xsnprintf(var, sizeof var, "%s,%ld,%d", socket_path, pid, idx);
-	environ_set(env, "TMUX", var);
-}
-
-void
-server_write_ready(struct client *c)
-{
-	if (c->flags & CLIENT_CONTROL)
-		return;
-	server_write_client(c, MSG_READY, NULL, 0);
-}
-
-int
-server_write_client(struct client *c, enum msgtype type, const void *buf,
-    size_t len)
-{
-	struct imsgbuf	*ibuf = &c->ibuf;
-	int              error;
-
-	if (c->flags & CLIENT_BAD)
-		return (-1);
-	log_debug("writing %d to client %d", type, c->ibuf.fd);
-	error = imsg_compose(ibuf, type, PROTOCOL_VERSION, -1, -1,
-	    (void *) buf, len);
-	if (error == 1)
-		server_update_event(c);
-	return (error == 1 ? 0 : -1);
-}
-
-void
-server_write_session(struct session *s, enum msgtype type, const void *buf,
-    size_t len)
-{
-	struct client	*c;
-	u_int		 i;
-
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c == NULL || c->session == NULL)
-			continue;
-		if (c->session == s)
-			server_write_client(c, type, buf, len);
-	}
+	environ_set(env, "TMUX", "%s,%ld,%u", socket_path, pid, idx);
 }
 
 void
@@ -104,12 +63,8 @@ void
 server_redraw_session(struct session *s)
 {
 	struct client	*c;
-	u_int		 i;
 
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c == NULL || c->session == NULL)
-			continue;
+	TAILQ_FOREACH(c, &clients, entry) {
 		if (c->session == s)
 			server_redraw_client(c);
 	}
@@ -132,12 +87,8 @@ void
 server_status_session(struct session *s)
 {
 	struct client	*c;
-	u_int		 i;
 
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c == NULL || c->session == NULL)
-			continue;
+	TAILQ_FOREACH(c, &clients, entry) {
 		if (c->session == s)
 			server_status_client(c);
 	}
@@ -160,13 +111,9 @@ void
 server_redraw_window(struct window *w)
 {
 	struct client	*c;
-	u_int		 i;
 
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c == NULL || c->session == NULL)
-			continue;
-		if (c->session->curw->window == w)
+	TAILQ_FOREACH(c, &clients, entry) {
+		if (c->session != NULL && c->session->curw->window == w)
 			server_redraw_client(c);
 	}
 	w->flags |= WINDOW_REDRAW;
@@ -176,13 +123,9 @@ void
 server_redraw_window_borders(struct window *w)
 {
 	struct client	*c;
-	u_int		 i;
 
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c == NULL || c->session == NULL)
-			continue;
-		if (c->session->curw->window == w)
+	TAILQ_FOREACH(c, &clients, entry) {
+		if (c->session != NULL && c->session->curw->window == w)
 			c->flags |= CLIENT_BORDERS;
 	}
 }
@@ -199,7 +142,7 @@ server_status_window(struct window *w)
 	 */
 
 	RB_FOREACH(s, sessions, &sessions) {
-		if (session_has(s, w) != NULL)
+		if (session_has(s, w))
 			server_status_session(s);
 	}
 }
@@ -208,13 +151,10 @@ void
 server_lock(void)
 {
 	struct client	*c;
-	u_int		 i;
 
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c == NULL || c->session == NULL)
-			continue;
-		server_lock_client(c);
+	TAILQ_FOREACH(c, &clients, entry) {
+		if (c->session != NULL)
+			server_lock_client(c);
 	}
 }
 
@@ -222,13 +162,10 @@ void
 server_lock_session(struct session *s)
 {
 	struct client	*c;
-	u_int		 i;
 
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c == NULL || c->session == NULL || c->session != s)
-			continue;
-		server_lock_client(c);
+	TAILQ_FOREACH(c, &clients, entry) {
+		if (c->session == s)
+			server_lock_client(c);
 	}
 }
 
@@ -243,7 +180,7 @@ server_lock_client(struct client *c)
 	if (c->flags & CLIENT_SUSPENDED)
 		return;
 
-	cmd = options_get_string(&c->session->options, "lock-command");
+	cmd = options_get_string(c->session->options, "lock-command");
 	if (strlen(cmd) + 1 > MAX_IMSGSIZE - IMSG_HEADER_SIZE)
 		return;
 
@@ -253,7 +190,7 @@ server_lock_client(struct client *c)
 	tty_raw(&c->tty, tty_term_string(c->tty.term, TTYC_E3));
 
 	c->flags |= CLIENT_SUSPENDED;
-	server_write_client(c, MSG_LOCK, cmd, strlen(cmd) + 1);
+	proc_send_s(c->peer, MSG_LOCK, cmd);
 }
 
 void
@@ -268,7 +205,7 @@ server_kill_window(struct window *w)
 		s = next_s;
 		next_s = RB_NEXT(sessions, &sessions, s);
 
-		if (session_has(s, w) == NULL)
+		if (!session_has(s, w))
 			continue;
 		server_unzoom_window(w);
 		while ((wl = winlink_find_by_window(&s->windows, w)) != NULL) {
@@ -279,7 +216,7 @@ server_kill_window(struct window *w)
 				server_redraw_session_group(s);
 		}
 
-		if (options_get_number(&s->options, "renumber-windows")) {
+		if (options_get_number(s->options, "renumber-windows")) {
 			if ((sg = session_group_find(s)) != NULL) {
 				TAILQ_FOREACH(target_s, &sg->sessions, gentry)
 					session_renumber_windows(target_s);
@@ -332,7 +269,7 @@ server_link_window(struct session *src, struct winlink *srcwl,
 	}
 
 	if (dstidx == -1)
-		dstidx = -1 - options_get_number(&dst->options, "base-index");
+		dstidx = -1 - options_get_number(dst->options, "base-index");
 	dstwl = session_attach(dst, srcwl->window, dstidx, cause);
 	if (dstwl == NULL)
 		return (-1);
@@ -354,12 +291,13 @@ server_unlink_window(struct session *s, struct winlink *wl)
 }
 
 void
-server_destroy_pane(struct window_pane *wp)
+server_destroy_pane(struct window_pane *wp, int hooks)
 {
 	struct window		*w = wp->window;
 	int			 old_fd;
 	struct screen_write_ctx	 ctx;
 	struct grid_cell	 gc;
+	struct cmd_find_state	 fs;
 
 	old_fd = wp->fd;
 	if (wp->fd != -1) {
@@ -371,7 +309,7 @@ server_destroy_pane(struct window_pane *wp)
 		wp->fd = -1;
 	}
 
-	if (options_get_number(&w->options, "remain-on-exit")) {
+	if (options_get_number(w->options, "remain-on-exit")) {
 		if (old_fd == -1)
 			return;
 		screen_write_start(&ctx, wp, &wp->base);
@@ -383,6 +321,9 @@ server_destroy_pane(struct window_pane *wp)
 		screen_write_puts(&ctx, &gc, "Pane is dead");
 		screen_write_stop(&ctx);
 		wp->flags |= PANE_REDRAW;
+
+		if (hooks && cmd_find_from_pane(&fs, wp) == 0)
+			hooks_run(hooks_get(fs.s), NULL, &fs, "pane-died");
 		return;
 	}
 
@@ -390,6 +331,9 @@ server_destroy_pane(struct window_pane *wp)
 	layout_close_pane(wp);
 	window_remove_pane(w, wp);
 
+	if (hooks && cmd_find_from_window(&fs, w) == 0)
+		hooks_run(hooks_get(fs.s), NULL, &fs, "pane-exited");
+
 	if (TAILQ_EMPTY(&w->panes))
 		server_kill_window(w);
 	else
@@ -433,16 +377,14 @@ server_destroy_session(struct session *s)
 {
 	struct client	*c;
 	struct session	*s_new;
-	u_int		 i;
 
-	if (!options_get_number(&s->options, "detach-on-destroy"))
+	if (!options_get_number(s->options, "detach-on-destroy"))
 		s_new = server_next_session(s);
 	else
 		s_new = NULL;
 
-	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
-		c = ARRAY_ITEM(&clients, i);
-		if (c == NULL || c->session != s)
+	TAILQ_FOREACH(c, &clients, entry) {
+		if (c->session != s)
 			continue;
 		if (s_new == NULL) {
 			c->session = NULL;
@@ -450,9 +392,13 @@ server_destroy_session(struct session *s)
 		} else {
 			c->last_session = NULL;
 			c->session = s_new;
+			server_client_set_key_table(c, NULL);
+			status_timer_start(c);
 			notify_attached_session_changed(c);
-			session_update_activity(s_new);
+			session_update_activity(s_new, NULL);
+			gettimeofday(&s_new->last_attached_time, NULL);
 			server_redraw_client(c);
+			alerts_check_session(s_new);
 		}
 	}
 	recalculate_sizes();
@@ -470,7 +416,7 @@ server_check_unattached(void)
 	RB_FOREACH(s, sessions, &sessions) {
 		if (!(s->flags & SESSION_UNATTACHED))
 			continue;
-		if (options_get_number (&s->options, "destroy-unattached"))
+		if (options_get_number (s->options, "destroy-unattached"))
 			session_destroy(s);
 	}
 }
@@ -481,7 +427,7 @@ server_set_identify(struct client *c)
 	struct timeval	tv;
 	int		delay;
 
-	delay = options_get_number(&c->session->options, "display-panes-time");
+	delay = options_get_number(c->session->options, "display-panes-time");
 	tv.tv_sec = delay / 1000;
 	tv.tv_usec = (delay % 1000) * 1000L;
 
@@ -496,81 +442,23 @@ server_set_identify(struct client *c)
 }
 
 void
-server_clear_identify(struct client *c)
+server_clear_identify(struct client *c, struct window_pane *wp)
 {
-	if (c->flags & CLIENT_IDENTIFY) {
-		c->flags &= ~CLIENT_IDENTIFY;
-		c->tty.flags &= ~(TTY_FREEZE|TTY_NOCURSOR);
-		server_redraw_client(c);
-	}
-}
-
-void
-server_callback_identify(unused int fd, unused short events, void *data)
-{
-	struct client	*c = data;
-
-	server_clear_identify(c);
-}
-
-void
-server_update_event(struct client *c)
-{
-	short	events;
-
-	events = 0;
-	if (!(c->flags & CLIENT_BAD))
-		events |= EV_READ;
-	if (c->ibuf.w.queued > 0)
-		events |= EV_WRITE;
-	if (event_initialized(&c->event))
-		event_del(&c->event);
-	event_set(&c->event, c->ibuf.fd, events, server_client_callback, c);
-	event_add(&c->event, NULL);
-}
-
-/* Push stdout to client if possible. */
-void
-server_push_stdout(struct client *c)
-{
-	struct msg_stdout_data data;
-	size_t                 size;
-
-	size = EVBUFFER_LENGTH(c->stdout_data);
-	if (size == 0)
+	if (~c->flags & CLIENT_IDENTIFY)
 		return;
-	if (size > sizeof data.data)
-		size = sizeof data.data;
+	c->flags &= ~CLIENT_IDENTIFY;
 
-	memcpy(data.data, EVBUFFER_DATA(c->stdout_data), size);
-	data.size = size;
+	if (c->identify_callback != NULL)
+		c->identify_callback(c, wp);
 
-	if (server_write_client(c, MSG_STDOUT, &data, sizeof data) == 0)
-		evbuffer_drain(c->stdout_data, size);
+	c->tty.flags &= ~(TTY_FREEZE|TTY_NOCURSOR);
+	server_redraw_client(c);
 }
 
-/* Push stderr to client if possible. */
 void
-server_push_stderr(struct client *c)
+server_callback_identify(__unused int fd, __unused short events, void *data)
 {
-	struct msg_stderr_data data;
-	size_t                 size;
-
-	if (c->stderr_data == c->stdout_data) {
-		server_push_stdout(c);
-		return;
-	}
-	size = EVBUFFER_LENGTH(c->stderr_data);
-	if (size == 0)
-		return;
-	if (size > sizeof data.data)
-		size = sizeof data.data;
-
-	memcpy(data.data, EVBUFFER_DATA(c->stderr_data), size);
-	data.size = size;
-
-	if (server_write_client(c, MSG_STDERR, &data, sizeof data) == 0)
-		evbuffer_drain(c->stderr_data, size);
+	server_clear_identify(data, NULL);
 }
 
 /* Set stdin callback. */
@@ -599,7 +487,7 @@ server_set_stdin_callback(struct client *c, void (*cb)(struct client *, int,
 	if (c->stdin_closed)
 		c->stdin_callback(c, 1, c->stdin_callback_data);
 
-	server_write_client(c, MSG_STDIN, NULL, 0);
+	proc_send(c->peer, MSG_STDIN, -1, NULL, 0);
 
 	return (0);
 }
@@ -607,7 +495,8 @@ server_set_stdin_callback(struct client *c, void (*cb)(struct client *, int,
 void
 server_unzoom_window(struct window *w)
 {
-	window_unzoom(w);
-	server_redraw_window(w);
-	server_status_window(w);
+	if (window_unzoom(w) == 0) {
+		server_redraw_window(w);
+		server_status_window(w);
+	}
 }
