@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -19,9 +19,12 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 
+#include <err.h>
 #include <errno.h>
 #include <event.h>
 #include <fcntl.h>
+#include <getopt.h>
+#include <langinfo.h>
 #include <locale.h>
 #include <pwd.h>
 #include <stdlib.h>
@@ -31,26 +34,17 @@
 
 #include "tmux.h"
 
-#if defined(DEBUG) && defined(__OpenBSD__)
-extern char	*malloc_options;
-#endif
+struct options	*global_options;	/* server options */
+struct options	*global_s_options;	/* session options */
+struct options	*global_w_options;	/* window options */
+struct environ	*global_environ;
+struct hooks	*global_hooks;
 
-struct options	 global_options;	/* server options */
-struct options	 global_s_options;	/* session options */
-struct options	 global_w_options;	/* window options */
-struct environ	 global_environ;
-
-char		*shell_cmd;
-int		 debug_level;
-time_t		 start_time;
-char		 socket_path[PATH_MAX];
+struct timeval	 start_time;
+const char	*socket_path;
 
 __dead void	 usage(void);
-char 		*makesocketpath(const char *);
-
-#ifndef HAVE___PROGNAME
-char      *__progname = (char *) "tmux";
-#endif
+static char	*make_label(const char *);
 
 __dead void
 usage(void)
@@ -58,22 +52,10 @@ usage(void)
 	fprintf(stderr,
 	    "usage: %s [-2CluvV] [-c shell-command] [-f file] [-L socket-name]\n"
 	    "            [-S socket-path] [command [flags]]\n",
-	    __progname);
+	    getprogname());
 	exit(1);
 }
 
-void
-logfile(const char *name)
-{
-	char	*path;
-
-	if (debug_level > 0) {
-		xasprintf(&path, "tmux-%s-%ld.log", name, (long) getpid());
-		log_open(path);
-		free(path);
-	}
-}
-
 const char *
 getshell(void)
 {
@@ -112,7 +94,7 @@ areshell(const char *shell)
 		ptr++;
 	else
 		ptr = shell;
-	progname = __progname;
+	progname = getprogname();
 	if (*progname == '-')
 		progname++;
 	if (strcmp(ptr, progname) == 0)
@@ -120,40 +102,48 @@ areshell(const char *shell)
 	return (0);
 }
 
-char *
-makesocketpath(const char *label)
+static char *
+make_label(const char *label)
 {
-	char		base[PATH_MAX], realbase[PATH_MAX], *path, *s;
-	struct stat	sb;
-	u_int		uid;
+	char		*base, resolved[PATH_MAX], *path, *s;
+	struct stat	 sb;
+	u_int		 uid;
+	int		 saved_errno;
+
+	if (label == NULL)
+		label = "default";
 
 	uid = getuid();
+
 	if ((s = getenv("TMUX_TMPDIR")) != NULL && *s != '\0')
-		xsnprintf(base, sizeof base, "%s/tmux-%u", s, uid);
-	else if ((s = getenv("TMPDIR")) != NULL && *s != '\0')
-		xsnprintf(base, sizeof base, "%s/tmux-%u", s, uid);
+		xasprintf(&base, "%s/tmux-%u", s, uid);
 	else
-		xsnprintf(base, sizeof base, "%s/tmux-%u", _PATH_TMP, uid);
+		xasprintf(&base, "%s/tmux-%u", _PATH_TMP, uid);
 
 	if (mkdir(base, S_IRWXU) != 0 && errno != EEXIST)
-		return (NULL);
+		goto fail;
 
 	if (lstat(base, &sb) != 0)
-		return (NULL);
+		goto fail;
 	if (!S_ISDIR(sb.st_mode)) {
 		errno = ENOTDIR;
-		return (NULL);
+		goto fail;
 	}
 	if (sb.st_uid != uid || (sb.st_mode & S_IRWXO) != 0) {
 		errno = EACCES;
-		return (NULL);
+		goto fail;
 	}
 
-	if (realpath(base, realbase) == NULL)
-		strlcpy(realbase, base, sizeof realbase);
-
-	xasprintf(&path, "%s/%s", realbase, label);
+	if (realpath(base, resolved) == NULL)
+		strlcpy(resolved, base, sizeof resolved);
+	xasprintf(&path, "%s/%s", resolved, label);
 	return (path);
+
+fail:
+	saved_errno = errno;
+	free(base);
+	errno = saved_errno;
+	return (NULL);
 }
 
 void
@@ -194,12 +184,17 @@ find_home(void)
 int
 main(int argc, char **argv)
 {
-	char	*s, *path, *label, **var, tmp[PATH_MAX];
-	int	 opt, flags, keys;
+	char		*path, *label, **var, tmp[PATH_MAX], *shellcmd = NULL;
+	const char	*s;
+	int		 opt, flags, keys;
 
-#if defined(DEBUG) && defined(__OpenBSD__)
-	malloc_options = (char *) "AFGJPX";
-#endif
+	if (setlocale(LC_CTYPE, "en_US.UTF-8") == NULL) {
+		if (setlocale(LC_CTYPE, "") == NULL)
+			errx(1, "invalid LC_ALL, LC_CTYPE or LANG");
+		s = nl_langinfo(CODESET);
+		if (strcasecmp(s, "UTF-8") != 0 && strcasecmp(s, "UTF8") != 0)
+			errx(1, "need UTF-8 locale (LC_CTYPE) but have %s", s);
+	}
 
 	setlocale(LC_TIME, "");
 	tzset();
@@ -216,8 +211,8 @@ main(int argc, char **argv)
 			flags |= CLIENT_256COLOURS;
 			break;
 		case 'c':
-			free(shell_cmd);
-			shell_cmd = xstrdup(optarg);
+			free(shellcmd);
+			shellcmd = xstrdup(optarg);
 			break;
 		case 'C':
 			if (flags & CLIENT_CONTROL)
@@ -226,7 +221,7 @@ main(int argc, char **argv)
 				flags |= CLIENT_CONTROL;
 			break;
 		case 'V':
-			printf("%s %s\n", __progname, VERSION);
+			printf("%s %s\n", getprogname(), VERSION);
 			exit(0);
 		case 'f':
 			set_cfg_file(optarg);
@@ -248,7 +243,7 @@ main(int argc, char **argv)
 			flags |= CLIENT_UTF8;
 			break;
 		case 'v':
-			debug_level++;
+			log_add_level();
 			break;
 		default:
 			usage();
@@ -257,49 +252,54 @@ main(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
-	if (shell_cmd != NULL && argc != 0)
+	if (shellcmd != NULL && argc != 0)
 		usage();
 
-	if (!(flags & CLIENT_UTF8)) {
-		/*
-		 * If the user has set whichever of LC_ALL, LC_CTYPE or LANG
-		 * exist (in that order) to contain UTF-8, it is a safe
-		 * assumption that either they are using a UTF-8 terminal, or
-		 * if not they know that output from UTF-8-capable programs may
-		 * be wrong.
-		 */
-		if ((s = getenv("LC_ALL")) == NULL || *s == '\0') {
-			if ((s = getenv("LC_CTYPE")) == NULL || *s == '\0')
-				s = getenv("LANG");
-		}
-		if (s != NULL && (strcasestr(s, "UTF-8") != NULL ||
-		    strcasestr(s, "UTF8") != NULL))
+#ifdef __OpenBSD__
+	if (pledge("stdio rpath wpath cpath flock fattr unix getpw sendfd "
+	    "recvfd proc exec tty ps", NULL) != 0)
+		err(1, "pledge");
+#endif
+
+	/*
+	 * tmux is a UTF-8 terminal, so if TMUX is set, assume UTF-8.
+	 * Otherwise, if the user has set LC_ALL, LC_CTYPE or LANG to contain
+	 * UTF-8, it is a safe assumption that either they are using a UTF-8
+	 * terminal, or if not they know that output from UTF-8-capable
+	 * programs may be wrong.
+	 */
+	if (getenv("TMUX") != NULL)
+		flags |= CLIENT_UTF8;
+	else {
+		s = getenv("LC_ALL");
+		if (s == NULL || *s == '\0')
+			s = getenv("LC_CTYPE");
+		if (s == NULL || *s == '\0')
+			s = getenv("LANG");
+		if (s == NULL || *s == '\0')
+			s = "";
+		if (strcasestr(s, "UTF-8") != NULL ||
+		    strcasestr(s, "UTF8") != NULL)
 			flags |= CLIENT_UTF8;
 	}
 
-	environ_init(&global_environ);
+	global_hooks = hooks_create(NULL);
+
+	global_environ = environ_create();
 	for (var = environ; *var != NULL; var++)
-		environ_put(&global_environ, *var);
+		environ_put(global_environ, *var);
 	if (getcwd(tmp, sizeof tmp) != NULL)
-		environ_set(&global_environ, "PWD", tmp);
+		environ_set(global_environ, "PWD", "%s", tmp);
 
-	options_init(&global_options, NULL);
-	options_table_populate_tree(server_options_table, &global_options);
+	global_options = options_create(NULL);
+	options_table_populate_tree(OPTIONS_TABLE_SERVER, global_options);
 
-	options_init(&global_s_options, NULL);
-	options_table_populate_tree(session_options_table, &global_s_options);
-	options_set_string(&global_s_options, "default-shell", "%s",
-	    getshell());
+	global_s_options = options_create(NULL);
+	options_table_populate_tree(OPTIONS_TABLE_SESSION, global_s_options);
+	options_set_string(global_s_options, "default-shell", "%s", getshell());
 
-	options_init(&global_w_options, NULL);
-	options_table_populate_tree(window_options_table, &global_w_options);
-
-	/* Enable UTF-8 if the first client is on UTF-8 terminal. */
-	if (flags & CLIENT_UTF8) {
-		options_set_number(&global_s_options, "status-utf8", 1);
-		options_set_number(&global_s_options, "mouse-utf8", 1);
-		options_set_number(&global_w_options, "utf8", 1);
-	}
+	global_w_options = options_create(NULL);
+	options_table_populate_tree(OPTIONS_TABLE_WINDOW, global_w_options);
 
 	/* Override keys to vi if VISUAL or EDITOR are set. */
 	if ((s = getenv("VISUAL")) != NULL || (s = getenv("EDITOR")) != NULL) {
@@ -309,52 +309,29 @@ main(int argc, char **argv)
 			keys = MODEKEY_VI;
 		else
 			keys = MODEKEY_EMACS;
-		options_set_number(&global_s_options, "status-keys", keys);
-		options_set_number(&global_w_options, "mode-keys", keys);
+		options_set_number(global_s_options, "status-keys", keys);
+		options_set_number(global_w_options, "mode-keys", keys);
 	}
 
 	/*
-	 * Figure out the socket path. If specified on the command-line with -S
-	 * or -L, use it, otherwise try $TMUX or assume -L default.
+	 * If socket is specified on the command-line with -S or -L, it is
+	 * used. Otherwise, $TMUX is checked and if that fails "default" is
+	 * used.
 	 */
-	if (path == NULL) {
-		/* If no -L, use the environment. */
-		if (label == NULL) {
-			s = getenv("TMUX");
-			if (s != NULL) {
-				path = xstrdup(s);
-				path[strcspn (path, ",")] = '\0';
-				if (*path == '\0') {
-					free(path);
-					label = xstrdup("default");
-				}
-			} else
-				label = xstrdup("default");
-		}
-
-		/* -L or default set. */
-		if (label != NULL) {
-			if ((path = makesocketpath(label)) == NULL) {
-				fprintf(stderr, "can't create socket: %s\n",
-				    strerror(errno));
-				exit(1);
-			}
+	if (path == NULL && label == NULL) {
+		s = getenv("TMUX");
+		if (s != NULL && *s != '\0' && *s != ',') {
+			path = xstrdup(s);
+			path[strcspn (path, ",")] = '\0';
 		}
 	}
-	free(label);
-
-	if (strlcpy(socket_path, path, sizeof socket_path) >=
-	    sizeof socket_path) {
-		fprintf(stderr, "socket path too long: %s\n", path);
+	if (path == NULL && (path = make_label(label)) == NULL) {
+		fprintf(stderr, "can't create socket: %s\n", strerror(errno));
 		exit(1);
 	}
-	free(path);
-
-#ifdef HAVE_SETPROCTITLE
-	/* Set process title. */
-	setproctitle("%s (%s)", __progname, socket_path);
-#endif
+	socket_path = path;
+	free(label);
 
 	/* Pass control to the client. */
-	exit(client_main(osdep_event_init(), argc, argv, flags));
+	exit(client_main(osdep_event_init(), argc, argv, flags, shellcmd));
 }
