@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -33,20 +33,22 @@
  * into a ternary tree.
  */
 
-void		tty_keys_add1(struct tty_key **, const char *, int);
-void		tty_keys_add(struct tty *, const char *, int);
-void		tty_keys_free1(struct tty_key *);
-struct tty_key *tty_keys_find1(
-		    struct tty_key *, const char *, size_t, size_t *);
-struct tty_key *tty_keys_find(struct tty *, const char *, size_t, size_t *);
-void		tty_keys_callback(int, short, void *);
-int		tty_keys_mouse(struct tty *, const char *, size_t, size_t *);
-int		tty_keys_device(struct tty *, const char *, size_t, size_t *);
+static void	tty_keys_add1(struct tty_key **, const char *, key_code);
+static void	tty_keys_add(struct tty *, const char *, key_code);
+static void	tty_keys_free1(struct tty_key *);
+static struct tty_key *tty_keys_find1(struct tty_key *, const char *, size_t,
+		    size_t *);
+static struct tty_key *tty_keys_find(struct tty *, const char *, size_t,
+		    size_t *);
+static int	tty_keys_next1(struct tty *, const char *, size_t, key_code *,
+		    size_t *, int);
+static void	tty_keys_callback(int, short, void *);
+static int	tty_keys_mouse(struct tty *, const char *, size_t, size_t *);
 
 /* Default raw keys. */
 struct tty_default_key_raw {
 	const char	       *string;
-	int	 	 	key;
+	key_code	 	key;
 };
 const struct tty_default_key_raw tty_default_raw_keys[] = {
 	/*
@@ -166,7 +168,7 @@ const struct tty_default_key_raw tty_default_raw_keys[] = {
 /* Default terminfo(5) keys. */
 struct tty_default_key_code {
 	enum tty_code_code	code;
-	int	 	 	key;
+	key_code	 	key;
 };
 const struct tty_default_key_code tty_default_code_keys[] = {
 	/* Function keys. */
@@ -317,8 +319,8 @@ const struct tty_default_key_code tty_default_code_keys[] = {
 };
 
 /* Add key to tree. */
-void
-tty_keys_add(struct tty *tty, const char *s, int key)
+static void
+tty_keys_add(struct tty *tty, const char *s, key_code key)
 {
 	struct tty_key	*tk;
 	size_t		 size;
@@ -326,17 +328,17 @@ tty_keys_add(struct tty *tty, const char *s, int key)
 
 	keystr = key_string_lookup_key(key);
 	if ((tk = tty_keys_find(tty, s, strlen(s), &size)) == NULL) {
-		log_debug("new key %s: 0x%x (%s)", s, key, keystr);
+		log_debug("new key %s: 0x%llx (%s)", s, key, keystr);
 		tty_keys_add1(&tty->key_tree, s, key);
 	} else {
-		log_debug("replacing key %s: 0x%x (%s)", s, key, keystr);
+		log_debug("replacing key %s: 0x%llx (%s)", s, key, keystr);
 		tk->key = key;
 	}
 }
 
 /* Add next node to the tree. */
-void
-tty_keys_add1(struct tty_key **tkp, const char *s, int key)
+static void
+tty_keys_add1(struct tty_key **tkp, const char *s, key_code key)
 {
 	struct tty_key	*tk;
 
@@ -345,7 +347,7 @@ tty_keys_add1(struct tty_key **tkp, const char *s, int key)
 	if (tk == NULL) {
 		tk = *tkp = xcalloc(1, sizeof *tk);
 		tk->ch = *s;
-		tk->key = KEYC_NONE;
+		tk->key = KEYC_UNKNOWN;
 	}
 
 	/* Find the next entry. */
@@ -410,7 +412,7 @@ tty_keys_free(struct tty *tty)
 }
 
 /* Free a single key. */
-void
+static void
 tty_keys_free1(struct tty_key *tk)
 {
 	if (tk->next != NULL)
@@ -423,7 +425,7 @@ tty_keys_free1(struct tty_key *tk)
 }
 
 /* Lookup a key in the tree. */
-struct tty_key *
+static struct tty_key *
 tty_keys_find(struct tty *tty, const char *buf, size_t len, size_t *size)
 {
 	*size = 0;
@@ -431,7 +433,7 @@ tty_keys_find(struct tty *tty, const char *buf, size_t len, size_t *size)
 }
 
 /* Find the next node. */
-struct tty_key *
+static struct tty_key *
 tty_keys_find1(struct tty_key *tk, const char *buf, size_t len, size_t *size)
 {
 	/* If the node is NULL, this is the end of the tree. No match. */
@@ -445,7 +447,7 @@ tty_keys_find1(struct tty_key *tk, const char *buf, size_t len, size_t *size)
 		(*size)++;
 
 		/* At the end of the string, return the current node. */
-		if (len == 0 || (tk->next == NULL && tk->key != KEYC_NONE))
+		if (len == 0 || (tk->next == NULL && tk->key != KEYC_UNKNOWN))
 			return (tk);
 
 		/* Move into the next tree for the following character. */
@@ -461,37 +463,79 @@ tty_keys_find1(struct tty_key *tk, const char *buf, size_t len, size_t *size)
 	return (tty_keys_find1(tk, buf, len, size));
 }
 
+/* Look up part of the next key. */
+static int
+tty_keys_next1(struct tty *tty, const char *buf, size_t len, key_code *key,
+    size_t *size, int expired)
+{
+	struct tty_key		*tk, *tk1;
+	struct utf8_data	 ud;
+	enum utf8_state		 more;
+	u_int			 i;
+	wchar_t			 wc;
+
+	log_debug("next key is %zu (%.*s) (expired=%d)", len, (int)len, buf,
+	    expired);
+
+	/* Is this a known key? */
+	tk = tty_keys_find(tty, buf, len, size);
+	if (tk != NULL && tk->key != KEYC_UNKNOWN) {
+		tk1 = tk;
+		do
+			log_debug("keys in list: %#llx", tk1->key);
+		while ((tk1 = tk1->next) != NULL);
+		if (tk->next != NULL && !expired)
+			return (1);
+		*key = tk->key;
+		return (0);
+	}
+
+	/* Is this valid UTF-8? */
+	more = utf8_open(&ud, (u_char)*buf);
+	if (more == UTF8_MORE) {
+		*size = ud.size;
+		if (len < ud.size) {
+			if (!expired)
+				return (1);
+			return (-1);
+		}
+		for (i = 1; i < ud.size; i++)
+			more = utf8_append(&ud, (u_char)buf[i]);
+		if (more != UTF8_DONE)
+			return (-1);
+
+		if (utf8_combine(&ud, &wc) != UTF8_DONE)
+			return (-1);
+		*key = wc;
+
+		log_debug("UTF-8 key %.*s %#llx", (int)ud.size, buf, *key);
+		return (0);
+	}
+
+	return (-1);
+}
+
 /*
  * Process at least one key in the buffer and invoke tty->key_callback. Return
  * 0 if there are no further keys, or 1 if there could be more in the buffer.
  */
-int
+key_code
 tty_keys_next(struct tty *tty)
 {
-	struct tty_key	*tk;
 	struct timeval	 tv;
 	const char	*buf;
 	size_t		 len, size;
 	cc_t		 bspace;
-	int		 key, delay, expired = 0;
+	int		 delay, expired = 0, n;
+	key_code	 key;
 
 	/* Get key buffer. */
 	buf = EVBUFFER_DATA(tty->event->input);
 	len = EVBUFFER_LENGTH(tty->event->input);
+
 	if (len == 0)
 		return (0);
-	log_debug("keys are %zu (%.*s)", len, (int) len, buf);
-
-	/* Is this device attributes response? */
-	switch (tty_keys_device(tty, buf, len, &size)) {
-	case 0:		/* yes */
-		key = KEYC_NONE;
-		goto complete_key;
-	case -1:	/* no, or not valid */
-		break;
-	case 1:		/* partial */
-		goto partial_key;
-	}
+	log_debug("keys are %zu (%.*s)", len, (int)len, buf);
 
 	/* Is this a mouse key press? */
 	switch (tty_keys_mouse(tty, buf, len, &size)) {
@@ -501,68 +545,55 @@ tty_keys_next(struct tty *tty)
 	case -1:	/* no, or not valid */
 		break;
 	case -2:	/* yes, but we don't care. */
+		key = KEYC_MOUSE;
 		goto discard_key;
 	case 1:		/* partial */
 		goto partial_key;
 	}
 
-	/* Look for matching key string and return if found. */
-	tk = tty_keys_find(tty, buf, len, &size);
-	if (tk != NULL) {
-		if (tk->next != NULL)
-			goto partial_key;
-		key = tk->key;
-		goto complete_key;
-	}
-
-	/* Try to parse a key with an xterm-style modifier. */
-	switch (xterm_keys_find(buf, len, &size, &key)) {
-	case 0:		/* found */
-		goto complete_key;
-	case -1:	/* not found */
-		break;
-	case 1:
-		goto partial_key;
-	}
-
 first_key:
-	/* Is this a meta key? */
-	if (len >= 2 && buf[0] == '\033') {
-		if (buf[1] != '\033') {
-			key = buf[1] | KEYC_ESCAPE;
-			size = 2;
-			goto complete_key;
-		}
-
-		tk = tty_keys_find(tty, buf + 1, len - 1, &size);
-		if (tk != NULL && (!expired || tk->next == NULL)) {
-			size++;	/* include escape */
-			if (tk->next != NULL)
-				goto partial_key;
-			key = tk->key;
-			if (key != KEYC_NONE)
-				key |= KEYC_ESCAPE;
+	/* Handle keys starting with escape. */
+	if (*buf == '\033') {
+		/* Look for a key without the escape. */
+		n = tty_keys_next1(tty, buf + 1, len - 1, &key, &size, expired);
+		if (n == 0) {	/* found */
+			key |= KEYC_ESCAPE;
+			size++;
 			goto complete_key;
 		}
+		if (n == 1)	/* partial */
+			goto partial_key;
 	}
 
-	/* No key found, take first. */
-	key = (u_char) *buf;
-	size = 1;
+	/* Try to lookup key. */
+	n = tty_keys_next1(tty, buf, len, &key, &size, expired);
+	if (n == 0)	/* found */
+		goto complete_key;
+	if (n == 1)
+		goto partial_key;
+
+	/* Is this an an xterm(1) key? */
+	n = xterm_keys_find(buf, len, &size, &key);
+	if (n == 0)
+		goto complete_key;
+	if (n == 1 && !expired)
+		goto partial_key;
 
 	/*
-	 * Check for backspace key using termios VERASE - the terminfo
-	 * kbs entry is extremely unreliable, so cannot be safely
-	 * used. termios should have a better idea.
+	 * At this point, we know the key is not partial (with or without
+	 * escape). So pass it through even if the timer has not expired.
 	 */
-	bspace = tty->tio.c_cc[VERASE];
-	if (bspace != _POSIX_VDISABLE && key == bspace)
-		key = KEYC_BSPACE;
-
+	if (*buf == '\033' && len >= 2) {
+		key = (u_char)buf[1] | KEYC_ESCAPE;
+		size = 2;
+	} else {
+		key = (u_char)buf[0];
+		size = 1;
+	}
 	goto complete_key;
 
 partial_key:
-	log_debug("partial key %.*s", (int) len, buf);
+	log_debug("partial key %.*s", (int)len, buf);
 
 	/* If timer is going, check for expiration. */
 	if (tty->flags & TTY_TIMER) {
@@ -575,7 +606,7 @@ partial_key:
 	}
 
 	/* Get the time period. */
-	delay = options_get_number(&global_options, "escape-time");
+	delay = options_get_number(global_options, "escape-time");
 	tv.tv_sec = delay / 1000;
 	tv.tv_usec = (delay % 1000) * 1000L;
 
@@ -589,7 +620,16 @@ partial_key:
 	return (0);
 
 complete_key:
-	log_debug("complete key %.*s %#x", (int) size, buf, key);
+	log_debug("complete key %.*s %#llx", (int)size, buf, key);
+
+	/*
+	 * Check for backspace key using termios VERASE - the terminfo
+	 * kbs entry is extremely unreliable, so cannot be safely
+	 * used. termios should have a better idea.
+	 */
+	bspace = tty->tio.c_cc[VERASE];
+	if (bspace != _POSIX_VDISABLE && (key & KEYC_MASK_KEY) == bspace)
+		key = (key & KEYC_MASK_MOD) | KEYC_BSPACE;
 
 	/* Remove data from buffer. */
 	evbuffer_drain(tty->event->input, size);
@@ -609,13 +649,13 @@ complete_key:
 	}
 
 	/* Fire the key. */
-	if (key != KEYC_NONE)
+	if (key != KEYC_UNKNOWN)
 		server_client_handle_key(tty->client, key);
 
 	return (1);
 
 discard_key:
-	log_debug("discard key %.*s %#x", (int) size, buf, key);
+	log_debug("discard key %.*s %#llx", (int)size, buf, key);
 
 	/* Remove data from buffer. */
 	evbuffer_drain(tty->event->input, size);
@@ -624,8 +664,8 @@ discard_key:
 }
 
 /* Key timer callback. */
-void
-tty_keys_callback(unused int fd, unused short events, void *data)
+static void
+tty_keys_callback(__unused int fd, __unused short events, void *data)
 {
 	struct tty	*tty = data;
 
@@ -639,13 +679,12 @@ tty_keys_callback(unused int fd, unused short events, void *data)
  * Handle mouse key input. Returns 0 for success, -1 for failure, 1 for partial
  * (probably a mouse sequence but need more data).
  */
-int
+static int
 tty_keys_mouse(struct tty *tty, const char *buf, size_t len, size_t *size)
 {
 	struct mouse_event	*m = &tty->mouse;
-	struct utf8_data	 utf8data;
-	u_int			 i, value, x, y, b, sgr, sgr_b, sgr_rel;
-	unsigned char		 c;
+	u_int			 i, x, y, b, sgr_b;
+	u_char			 sgr_type, c;
 
 	/*
 	 * Standard mouse sequences are \033[M followed by three characters
@@ -661,7 +700,8 @@ tty_keys_mouse(struct tty *tty, const char *buf, size_t len, size_t *size)
 	 */
 
 	*size = 0;
-	x = y = b = sgr = sgr_b = sgr_rel = 0;
+	x = y = b = sgr_b = 0;
+	sgr_type = ' ';
 
 	/* First two bytes are always \033[. */
 	if (buf[0] != '\033')
@@ -674,8 +714,8 @@ tty_keys_mouse(struct tty *tty, const char *buf, size_t len, size_t *size)
 		return (1);
 
 	/*
-	 * Third byte is M in old standard and UTF-8 extension, < in SGR
-	 * extension.
+	 * Third byte is M in old standard (and UTF-8 extension which we do not
+	 * support), < in SGR extension.
 	 */
 	if (buf[2] == 'M') {
 		/* Read the three inputs. */
@@ -683,32 +723,15 @@ tty_keys_mouse(struct tty *tty, const char *buf, size_t len, size_t *size)
 		for (i = 0; i < 3; i++) {
 			if (len <= *size)
 				return (1);
-
-			if (tty->mode & MODE_MOUSE_UTF8) {
-				if (utf8_open(&utf8data, buf[*size])) {
-					if (utf8data.size != 2)
-						return (-1);
-					(*size)++;
-					if (len <= *size)
-						return (1);
-					utf8_append(&utf8data, buf[*size]);
-					value = utf8_combine(&utf8data);
-				} else
-					value = (u_char) buf[*size];
-				(*size)++;
-			} else {
-				value = (u_char) buf[*size];
-				(*size)++;
-			}
-
+			c = (u_char)buf[(*size)++];
 			if (i == 0)
-				b = value;
+				b = c;
 			else if (i == 1)
-				x = value;
+				x = c;
 			else
-				y = value;
+				y = c;
 		}
-		log_debug("mouse input: %.*s", (int) *size, buf);
+		log_debug("mouse input: %.*s", (int)*size, buf);
 
 		/* Check and return the mouse input. */
 		if (b < 32)
@@ -748,22 +771,26 @@ tty_keys_mouse(struct tty *tty, const char *buf, size_t len, size_t *size)
 		while (1) {
 			if (len <= *size)
 				return (1);
-			c = (u_char) buf[(*size)++];
+			c = (u_char)buf[(*size)++];
 			if (c == 'M' || c == 'm')
 				break;
 			if (c < '0' || c > '9')
 				return (-1);
 			y = 10 * y + (c - '0');
 		}
-		log_debug("mouse input (sgr): %.*s", (int) *size, buf);
+		log_debug("mouse input (SGR): %.*s", (int)*size, buf);
 
 		/* Check and return the mouse input. */
 		if (x < 1 || y < 1)
 			return (-1);
 		x--;
 		y--;
-		sgr = 1;
-		sgr_rel = (c == 'm');
+		b = sgr_b;
+
+		/* Type is M for press, m for release. */
+		sgr_type = c;
+		if (sgr_type == 'm')
+			b |= 3;
 
 		/*
 		 * Some terminals (like PuTTY 0.63) mistakenly send
@@ -771,121 +798,20 @@ tty_keys_mouse(struct tty *tty, const char *buf, size_t len, size_t *size)
 		 * Discard it before it reaches any program running inside
 		 * tmux.
 		 */
-		if (sgr_rel && (sgr_b & 64))
+		if (sgr_type == 'm' && (sgr_b & 64))
 		    return (-2);
-
-		/* Figure out what b would be in old format. */
-		b = sgr_b;
-		if (sgr_rel)
-			b |= 3;
 	} else
 		return (-1);
 
-	/* Fill in mouse structure. */
-	if (~m->event & MOUSE_EVENT_WHEEL) {
-		m->lx = m->x;
-		m->ly = m->y;
-	}
-	m->xb = b;
-	m->sgr = sgr;
-	m->sgr_xb = sgr_b;
-	m->sgr_rel = sgr_rel;
+	/* Fill mouse event. */
+	m->lx = m->x;
 	m->x = x;
+	m->ly = m->y;
 	m->y = y;
-	if (b & MOUSE_MASK_WHEEL) {
-		if (b & MOUSE_MASK_SHIFT)
-			m->scroll = 1;
-		else
-			m->scroll = MOUSE_WHEEL_SCALE;
-		if (b & MOUSE_MASK_META)
-			m->scroll *= MOUSE_WHEEL_SCALE;
-		if (b & MOUSE_MASK_CTRL)
-			m->scroll *= MOUSE_WHEEL_SCALE;
-
-		b &= MOUSE_MASK_BUTTONS;
-		if (b == 0)
-			m->wheel = MOUSE_WHEEL_UP;
-		else if (b == 1)
-			m->wheel = MOUSE_WHEEL_DOWN;
-		m->event = MOUSE_EVENT_WHEEL;
-
-		m->button = 3;
-	} else if ((b & MOUSE_MASK_BUTTONS) == 3) {
-		if (~m->event & MOUSE_EVENT_DRAG && x == m->sx && y == m->sy) {
-			m->event = MOUSE_EVENT_CLICK;
-			m->clicks = (m->clicks + 1) % 3;
-		} else
-			m->event = MOUSE_EVENT_DRAG;
-		m->event |= MOUSE_EVENT_UP;
-	} else {
-		if (b & MOUSE_MASK_DRAG)
-			m->event = MOUSE_EVENT_DRAG;
-		else {
-			m->event = MOUSE_EVENT_DOWN;
-			if (x != m->sx || y != m->sy)
-				m->clicks = 0;
-		}
-		m->button = (b & MOUSE_MASK_BUTTONS);
-	}
-	m->sx = x;
-	m->sy = y;
-
-	return (0);
-}
-
-/*
- * Handle device attributes input. Returns 0 for success, -1 for failure, 1 for
- * partial.
- */
-int
-tty_keys_device(struct tty *tty, const char *buf, size_t len, size_t *size)
-{
-	u_int i, class;
-	char  tmp[64], *endptr;
-
-	/*
-	 * Primary device attributes are \033[?a;b and secondary are
-	 * \033[>a;b;c.
-	 */
-
-	*size = 0;
-
-	/* First three bytes are always \033[?. */
-	if (buf[0] != '\033')
-		return (-1);
-	if (len == 1)
-		return (1);
-	if (buf[1] != '[')
-		return (-1);
-	if (len == 2)
-		return (1);
-	if (buf[2] != '>' && buf[2] != '?')
-		return (-1);
-	if (len == 3)
-		return (1);
-
-	/* Copy the rest up to a 'c'. */
-	for (i = 0; i < (sizeof tmp) - 1 && buf[3 + i] != 'c'; i++) {
-		if (3 + i == len)
-			return (1);
-		tmp[i] = buf[3 + i];
-	}
-	if (i == (sizeof tmp) - 1)
-		return (-1);
-	tmp[i] = '\0';
-	*size = 4 + i;
-
-	/* Only primary is of interest. */
-	if (buf[2] != '?')
-		return (0);
-
-	/* Convert service class. */
-	class = strtoul(tmp, &endptr, 10);
-	if (*endptr != ';')
-		class = 0;
-
-	log_debug("received service class %u", class);
-	tty_set_class(tty, class);
+	m->lb = m->b;
+	m->b = b;
+	m->sgr_type = sgr_type;
+	m->sgr_b = sgr_b;
 
 	return (0);
 }
