@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2009 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2009 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -22,24 +22,23 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include "array.h"
 #include "tmux.h"
 
 struct screen *window_choose_init(struct window_pane *);
 void	window_choose_free(struct window_pane *);
 void	window_choose_resize(struct window_pane *, u_int, u_int);
 void	window_choose_key(struct window_pane *, struct client *,
-	    struct session *, int, struct mouse_event *);
+	    struct session *, key_code, struct mouse_event *);
 
 void	window_choose_default_callback(struct window_choose_data *);
 struct window_choose_mode_item *window_choose_get_item(struct window_pane *,
-	    int, struct mouse_event *);
+	    key_code, struct mouse_event *);
 
-void	window_choose_fire_callback(
-	    struct window_pane *, struct window_choose_data *);
+void	window_choose_fire_callback(struct window_pane *,
+	    struct window_choose_data *);
 void	window_choose_redraw_screen(struct window_pane *);
-void	window_choose_write_line(
-	    struct window_pane *, struct screen_write_ctx *, u_int);
+void	window_choose_write_line(struct window_pane *,
+	    struct screen_write_ctx *, u_int);
 
 void	window_choose_scroll_up(struct window_pane *);
 void	window_choose_scroll_down(struct window_pane *);
@@ -72,8 +71,11 @@ struct window_choose_mode_data {
 
 	struct mode_key_data	mdata;
 
-	ARRAY_DECL(, struct window_choose_mode_item) list;
-	ARRAY_DECL(, struct window_choose_mode_item) old_list;
+	struct window_choose_mode_item *list;
+	u_int			list_size;
+	struct window_choose_mode_item *old_list;
+	u_int			old_list_size;
+
 	int			width;
 	u_int			top;
 	u_int			selected;
@@ -86,9 +88,9 @@ struct window_choose_mode_data {
 
 void	window_choose_free1(struct window_choose_mode_data *);
 int     window_choose_key_index(struct window_choose_mode_data *, u_int);
-int     window_choose_index_key(struct window_choose_mode_data *, int);
+int     window_choose_index_key(struct window_choose_mode_data *, key_code);
 void	window_choose_prompt_input(enum window_choose_input_type,
-	    const char *, struct window_pane *, int);
+	    const char *, struct window_pane *, key_code);
 void	window_choose_reset_top(struct window_pane *, u_int);
 
 void
@@ -96,17 +98,18 @@ window_choose_add(struct window_pane *wp, struct window_choose_data *wcd)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
 	struct window_choose_mode_item	*item;
-	char				 tmp[10];
+	char				 tmp[11];
 
-	ARRAY_EXPAND(&data->list, 1);
-	item = &ARRAY_LAST(&data->list);
+	data->list = xreallocarray(data->list, data->list_size + 1,
+	    sizeof *data->list);
+	item = &data->list[data->list_size++];
 
 	item->name = format_expand(wcd->ft, wcd->ft_template);
 	item->wcd = wcd;
-	item->pos = ARRAY_LENGTH(&data->list) - 1;
+	item->pos = data->list_size - 1;
 	item->state = 0;
 
-	data->width = xsnprintf(tmp, sizeof tmp , "%d", item->pos);
+	data->width = xsnprintf(tmp, sizeof tmp, "%d", item->pos);
 }
 
 void
@@ -136,12 +139,18 @@ window_choose_ready(struct window_pane *wp, u_int cur,
     void (*callbackfn)(struct window_choose_data *))
 {
 	struct window_choose_mode_data	*data = wp->modedata;
+	u_int				 size;
 
 	data->callbackfn = callbackfn;
 	if (data->callbackfn == NULL)
 		data->callbackfn = window_choose_default_callback;
 
-	ARRAY_CONCAT(&data->old_list, &data->list);
+	size = data->old_list_size;
+	data->old_list_size += data->list_size;
+	data->old_list = xreallocarray(data->old_list, data->old_list_size,
+	    sizeof *data->old_list);
+	memcpy(data->old_list + size, data->list, data->list_size *
+	    sizeof *data->list);
 
 	window_choose_set_current(wp, cur);
 	window_choose_collapse_all(wp);
@@ -154,22 +163,26 @@ window_choose_init(struct window_pane *wp)
 	struct screen			*s;
 	int				 keys;
 
-	wp->modedata = data = xmalloc(sizeof *data);
+	wp->modedata = data = xcalloc(1, sizeof *data);
 
 	data->callbackfn = NULL;
 	data->input_type = WINDOW_CHOOSE_NORMAL;
 	data->input_str = xstrdup("");
 	data->input_prompt = NULL;
 
-	ARRAY_INIT(&data->list);
-	ARRAY_INIT(&data->old_list);
+	data->list = NULL;
+	data->list_size = 0;
+
+	data->old_list = NULL;
+	data->old_list_size = 0;
+
 	data->top = 0;
 
 	s = &data->screen;
 	screen_init(s, screen_size_x(&wp->base), screen_size_y(&wp->base), 0);
 	s->mode &= ~MODE_CURSOR;
 
-	keys = options_get_number(&wp->window->options, "mode-keys");
+	keys = options_get_number(wp->window->options, "mode-keys");
 	if (keys == MODEKEY_EMACS)
 		mode_key_init(&data->mdata, &mode_key_tree_emacs_choice);
 	else
@@ -186,7 +199,7 @@ window_choose_data_create(int type, struct client *c, struct session *s)
 	wcd = xmalloc(sizeof *wcd);
 	wcd->type = type;
 
-	wcd->ft = format_create();
+	wcd->ft = format_create(NULL, 0);
 	wcd->ft_template = NULL;
 
 	wcd->command = NULL;
@@ -274,13 +287,14 @@ window_choose_free1(struct window_choose_mode_data *data)
 	if (data == NULL)
 		return;
 
-	for (i = 0; i < ARRAY_LENGTH(&data->old_list); i++) {
-		item = &ARRAY_ITEM(&data->old_list, i);
+	for (i = 0; i < data->old_list_size; i++) {
+		item = &data->old_list[i];
 		window_choose_data_free(item->wcd);
 		free(item->name);
 	}
-	ARRAY_FREE(&data->list);
-	ARRAY_FREE(&data->old_list);
+	free(data->list);
+	free(data->old_list);
+
 	free(data->input_str);
 
 	screen_free(&data->screen);
@@ -299,8 +313,8 @@ window_choose_resize(struct window_pane *wp, u_int sx, u_int sy)
 }
 
 void
-window_choose_fire_callback(
-    struct window_pane *wp, struct window_choose_data *wcd)
+window_choose_fire_callback(struct window_pane *wp,
+    struct window_choose_data *wcd)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
 
@@ -314,7 +328,7 @@ window_choose_fire_callback(
 
 void
 window_choose_prompt_input(enum window_choose_input_type input_type,
-    const char *prompt, struct window_pane *wp, int key)
+    const char *prompt, struct window_pane *wp, key_code key)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
 	size_t				 input_len;
@@ -334,14 +348,11 @@ void
 window_choose_collapse(struct window_pane *wp, struct session *s, u_int pos)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
-	struct window_choose_mode_item	*item, *chosen;
+	struct window_choose_mode_item	*item, *chosen, *copy = NULL;
 	struct window_choose_data	*wcd;
-	u_int				 i;
+	u_int				 i, copy_size = 0;
 
-	ARRAY_DECL(, struct window_choose_mode_item) list_copy;
-	ARRAY_INIT(&list_copy);
-
-	chosen = &ARRAY_ITEM(&data->list, pos);
+	chosen = &data->list[pos];
 	chosen->state &= ~TREE_EXPANDED;
 
 	/*
@@ -349,15 +360,19 @@ window_choose_collapse(struct window_pane *wp, struct session *s, u_int pos)
 	 * assign the actual result we want to render and copy the new one over
 	 * the top of it.
 	 */
-	for (i = 0; i < ARRAY_LENGTH(&data->list); i++) {
-		item = &ARRAY_ITEM(&data->list, i);
+	for (i = 0; i < data->list_size; i++) {
+		item = &data->list[i];
 		wcd = item->wcd;
 
 		if (s == wcd->tree_session) {
 			/* We only show the session when collapsed. */
 			if (wcd->type & TREE_SESSION) {
 				item->state &= ~TREE_EXPANDED;
-				ARRAY_ADD(&list_copy, *item);
+
+				copy = xreallocarray(copy, copy_size + 1,
+				    sizeof *copy);
+				memcpy(&copy[copy_size], item, sizeof *copy);
+				copy_size++;
 
 				/*
 				 * Update the selection to this session item so
@@ -366,14 +381,17 @@ window_choose_collapse(struct window_pane *wp, struct session *s, u_int pos)
 				 */
 				data->selected = i;
 			}
-		} else
-			ARRAY_ADD(&list_copy, ARRAY_ITEM(&data->list, i));
+		} else {
+			copy = xreallocarray(copy, copy_size + 1, sizeof *copy);
+			memcpy(&copy[copy_size], item, sizeof *copy);
+			copy_size++;
+		}
 	}
 
-	if (!ARRAY_EMPTY(&list_copy)) {
-		ARRAY_FREE(&data->list);
-		ARRAY_CONCAT(&data->list, &list_copy);
-		ARRAY_FREE(&list_copy);
+	if (copy_size != 0) {
+		free(data->list);
+		data->list = copy;
+		data->list_size = copy_size;
 	}
 }
 
@@ -386,14 +404,14 @@ window_choose_collapse_all(struct window_pane *wp)
 	struct session			*s, *chosen;
 	u_int				 i;
 
-	chosen = ARRAY_ITEM(&data->list, data->selected).wcd->start_session;
+	chosen = data->list[data->selected].wcd->start_session;
 
 	RB_FOREACH(s, sessions, &sessions)
 		window_choose_collapse(wp, s, data->selected);
 
 	/* Reset the selection back to the starting session. */
-	for (i = 0; i < ARRAY_LENGTH(&data->list); i++) {
-		item = &ARRAY_ITEM(&data->list, i);
+	for (i = 0; i < data->list_size; i++) {
+		item = &data->list[i];
 
 		if (chosen != item->wcd->tree_session)
 			continue;
@@ -414,8 +432,8 @@ window_choose_expand_all(struct window_pane *wp)
 	u_int				 i;
 
 	RB_FOREACH(s, sessions, &sessions) {
-		for (i = 0; i < ARRAY_LENGTH(&data->list); i++) {
-			item = &ARRAY_ITEM(&data->list, i);
+		for (i = 0; i < data->list_size; i++) {
+			item = &data->list[i];
 
 			if (s != item->wcd->tree_session)
 				continue;
@@ -436,8 +454,8 @@ window_choose_expand(struct window_pane *wp, struct session *s, u_int pos)
 	struct window_choose_data	*wcd;
 	u_int				 i, items;
 
-	chosen = &ARRAY_ITEM(&data->list, pos);
-	items = ARRAY_LENGTH(&data->old_list) - 1;
+	chosen = &data->list[pos];
+	items = data->old_list_size - 1;
 
 	/* It's not possible to expand anything other than sessions. */
 	if (!(chosen->wcd->type & TREE_SESSION))
@@ -456,7 +474,7 @@ window_choose_expand(struct window_pane *wp, struct session *s, u_int pos)
 	 * to expand.
 	 */
 	for (i = items; i > 0; i--) {
-		item = &ARRAY_ITEM(&data->old_list, i);
+		item = &data->old_list[i];
 		item->state |= TREE_EXPANDED;
 		wcd = item->wcd;
 
@@ -473,16 +491,27 @@ window_choose_expand(struct window_pane *wp, struct session *s, u_int pos)
 				 * entries in order *AFTER* the selected
 				 * session.
 				 */
-				if (pos < i ) {
-					ARRAY_INSERT(&data->list,
-					    pos + 1,
-					    ARRAY_ITEM(&data->old_list,
-					    i));
+				if (pos < i) {
+					data->list = xreallocarray(data->list,
+					    data->list_size + 1,
+					    sizeof *data->list);
+					memmove(&data->list[pos + 2],
+					    &data->list[pos + 1],
+					    (data->list_size - (pos + 1)) *
+					    sizeof *data->list);
+					memcpy(&data->list[pos + 1],
+					    &data->old_list[i],
+					    sizeof *data->list);
+					data->list_size++;
 				} else {
 					/* Ran out of room, add to the end. */
-					ARRAY_ADD(&data->list,
-					    ARRAY_ITEM(&data->old_list,
-					    i));
+					data->list = xreallocarray(data->list,
+					    data->list_size + 1,
+					    sizeof *data->list);
+					memcpy(&data->list[data->list_size],
+					    &data->old_list[i],
+					    sizeof *data->list);
+					data->list_size++;
 				}
 			}
 		}
@@ -490,26 +519,27 @@ window_choose_expand(struct window_pane *wp, struct session *s, u_int pos)
 }
 
 struct window_choose_mode_item *
-window_choose_get_item(struct window_pane *wp, int key, struct mouse_event *m)
+window_choose_get_item(struct window_pane *wp, key_code key,
+    struct mouse_event *m)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
 	u_int				 x, y, idx;
 
 	if (!KEYC_IS_MOUSE(key))
-		return (&ARRAY_ITEM(&data->list, data->selected));
+		return (&data->list[data->selected]);
 
 	if (cmd_mouse_at(wp, m, &x, &y, 0) != 0)
 		return (NULL);
 
 	idx = data->top + y;
-	if (idx >= ARRAY_LENGTH(&data->list))
+	if (idx >= data->list_size)
 		return (NULL);
-	return (&ARRAY_ITEM(&data->list, idx));
+	return (&data->list[idx]);
 }
 
 void
-window_choose_key(struct window_pane *wp, unused struct client *c,
-    unused struct session *sess, int key, struct mouse_event *m)
+window_choose_key(struct window_pane *wp, __unused struct client *c,
+    __unused struct session *sess, key_code key, struct mouse_event *m)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
 	struct screen			*s = &data->screen;
@@ -519,10 +549,10 @@ window_choose_key(struct window_pane *wp, unused struct client *c,
 	u_int				 items, n;
 	int				 idx;
 
-	items = ARRAY_LENGTH(&data->list);
+	items = data->list_size;
 
 	if (data->input_type == WINDOW_CHOOSE_GOTO_ITEM) {
-		switch (mode_key_lookup(&data->mdata, key, NULL)) {
+		switch (mode_key_lookup(&data->mdata, key, NULL, NULL)) {
 		case MODEKEYCHOICE_CANCEL:
 			data->input_type = WINDOW_CHOOSE_NORMAL;
 			window_choose_redraw_screen(wp);
@@ -534,8 +564,7 @@ window_choose_key(struct window_pane *wp, unused struct client *c,
 				window_choose_redraw_screen(wp);
 				break;
 			}
-			item = &ARRAY_ITEM(&data->list, n);
-			window_choose_fire_callback(wp, item->wcd);
+			window_choose_fire_callback(wp, data->list[n].wcd);
 			break;
 		case MODEKEYCHOICE_BACKSPACE:
 			input_len = strlen(data->input_str);
@@ -553,7 +582,7 @@ window_choose_key(struct window_pane *wp, unused struct client *c,
 		return;
 	}
 
-	switch (mode_key_lookup(&data->mdata, key, NULL)) {
+	switch (mode_key_lookup(&data->mdata, key, NULL, NULL)) {
 	case MODEKEYCHOICE_CANCEL:
 		window_choose_fire_callback(wp, NULL);
 		break;
@@ -613,10 +642,10 @@ window_choose_key(struct window_pane *wp, unused struct client *c,
 			window_choose_scroll_up(wp);
 		else {
 			screen_write_start(&ctx, wp, NULL);
-			window_choose_write_line(
-			    wp, &ctx, data->selected - data->top);
-			window_choose_write_line(
-			    wp, &ctx, data->selected + 1 - data->top);
+			window_choose_write_line(wp, &ctx,
+			    data->selected - data->top);
+			window_choose_write_line(wp, &ctx,
+			    data->selected + 1 - data->top);
 			screen_write_stop(&ctx);
 		}
 		break;
@@ -633,10 +662,10 @@ window_choose_key(struct window_pane *wp, unused struct client *c,
 
 		if (data->selected < data->top + screen_size_y(s)) {
 			screen_write_start(&ctx, wp, NULL);
-			window_choose_write_line(
-			    wp, &ctx, data->selected - data->top);
-			window_choose_write_line(
-			    wp, &ctx, data->selected - 1 - data->top);
+			window_choose_write_line(wp, &ctx,
+			    data->selected - data->top);
+			window_choose_write_line(wp, &ctx,
+			    data->selected - 1 - data->top);
 			screen_write_stop(&ctx);
 		} else
 			window_choose_scroll_down(wp);
@@ -648,8 +677,8 @@ window_choose_key(struct window_pane *wp, unused struct client *c,
 			data->selected--;
 			window_choose_scroll_up(wp);
 			screen_write_start(&ctx, wp, NULL);
-			window_choose_write_line(
-			    wp, &ctx, screen_size_y(s) - 1);
+			window_choose_write_line(wp, &ctx,
+			    screen_size_y(s) - 1);
 			screen_write_stop(&ctx);
 		} else
 			window_choose_scroll_up(wp);
@@ -732,41 +761,38 @@ window_choose_key(struct window_pane *wp, unused struct client *c,
 		break;
 	default:
 		idx = window_choose_index_key(data, key);
-		if (idx < 0 || (u_int) idx >= ARRAY_LENGTH(&data->list))
+		if (idx < 0 || (u_int) idx >= data->list_size)
 			break;
 		data->selected = idx;
-
-		item = &ARRAY_ITEM(&data->list, data->selected);
-		window_choose_fire_callback(wp, item->wcd);
+		window_choose_fire_callback(wp, data->list[idx].wcd);
 		break;
 	}
 }
 
 void
-window_choose_write_line(
-    struct window_pane *wp, struct screen_write_ctx *ctx, u_int py)
+window_choose_write_line(struct window_pane *wp, struct screen_write_ctx *ctx,
+    u_int py)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
 	struct window_choose_mode_item	*item;
-	struct options			*oo = &wp->window->options;
+	struct options			*oo = wp->window->options;
 	struct screen			*s = &data->screen;
 	struct grid_cell		 gc;
 	size_t				 last, xoff = 0;
 	char				 hdr[32], label[32];
-	int				 utf8flag, key;
+	int				 key;
 
 	if (data->callbackfn == NULL)
 		fatalx("called before callback assigned");
 
 	last = screen_size_y(s) - 1;
-	utf8flag = options_get_number(&wp->window->options, "utf8");
 	memcpy(&gc, &grid_default_cell, sizeof gc);
 	if (data->selected == data->top + py)
 		style_apply(&gc, oo, "mode-style");
 
 	screen_write_cursormove(ctx, 0, py);
-	if (data->top + py  < ARRAY_LENGTH(&data->list)) {
-		item = &ARRAY_ITEM(&data->list, data->top + py);
+	if (data->top + py  < data->list_size) {
+		item = &data->list[data->top + py];
 		if (item->wcd->wl != NULL &&
 		    item->wcd->wl->flags & WINLINK_ALERTFLAGS)
 			gc.attr |= GRID_ATTR_BRIGHT;
@@ -776,7 +802,7 @@ window_choose_write_line(
 			xsnprintf(label, sizeof label, "(%c)", key);
 		else
 			xsnprintf(label, sizeof label, "(%d)", item->pos);
-		screen_write_nputs(ctx, screen_size_x(s) - 1, &gc, utf8flag,
+		screen_write_nputs(ctx, screen_size_x(s) - 1, &gc,
 		    "%*s %s %s", data->width + 2, label,
 		    /*
 		     * Add indication to tree if necessary about whether it's
@@ -811,7 +837,7 @@ window_choose_key_index(struct window_choose_mode_data *data, u_int idx)
 	int			mkey;
 
 	for (ptr = keys; *ptr != '\0'; ptr++) {
-		mkey = mode_key_lookup(&data->mdata, *ptr, NULL);
+		mkey = mode_key_lookup(&data->mdata, *ptr, NULL, NULL);
 		if (mkey != MODEKEY_NONE && mkey != MODEKEY_OTHER)
 			continue;
 		if (idx-- == 0)
@@ -821,7 +847,7 @@ window_choose_key_index(struct window_choose_mode_data *data, u_int idx)
 }
 
 int
-window_choose_index_key(struct window_choose_mode_data *data, int key)
+window_choose_index_key(struct window_choose_mode_data *data, key_code key)
 {
 	static const char	keys[] = "0123456789"
 	                                 "abcdefghijklmnopqrstuvwxyz"
@@ -831,10 +857,10 @@ window_choose_index_key(struct window_choose_mode_data *data, int key)
 	u_int			idx = 0;
 
 	for (ptr = keys; *ptr != '\0'; ptr++) {
-		mkey = mode_key_lookup(&data->mdata, *ptr, NULL);
+		mkey = mode_key_lookup(&data->mdata, *ptr, NULL, NULL);
 		if (mkey != MODEKEY_NONE && mkey != MODEKEY_OTHER)
 			continue;
-		if (key == *ptr)
+		if (key == (key_code)*ptr)
 			return (idx);
 		idx++;
 	}
@@ -881,7 +907,7 @@ window_choose_scroll_down(struct window_pane *wp)
 	struct screen			*s = &data->screen;
 	struct screen_write_ctx		 ctx;
 
-	if (data->top >= ARRAY_LENGTH(&data->list))
+	if (data->top >= data->list_size)
 		return;
 	data->top++;
 
